"""
Management of Docker volumes

.. versionadded:: 2017.7.0

:depends: docker_ Python module

.. note::
    Older releases of the Python bindings for Docker were called docker-py_ in
    PyPI. All releases of docker_, and releases of docker-py_ >= 1.6.0 are
    supported. These python bindings can easily be installed using
    :py:func:`pip.install <salt.modules.pip.install>`:

    .. code-block:: bash

        salt myminion pip.install docker

    To upgrade from docker-py_ to docker_, you must first uninstall docker-py_,
    and then install docker_:

    .. code-block:: bash

        salt myminion pip.uninstall docker-py
        salt myminion pip.install docker

.. _docker: https://pypi.python.org/pypi/docker
.. _docker-py: https://pypi.python.org/pypi/docker-py

These states were moved from the :mod:`docker <salt.states.docker>` state
module (formerly called **dockerng**) in the 2017.7.0 release.
"""

import logging

import salt.utils.data

# Enable proper logging
log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Define the module's virtual name
__virtualname__ = "docker_volume"
__virtual_aliases__ = ("moby_volume",)


def __virtual__():
    """
    Only load if the docker execution module is available
    """
    if "docker.version" in __salt__:
        return __virtualname__
    return (False, __salt__.missing_fun_string("docker.version"))


def _find_volume(name):
    """
    Find volume by name on minion
    """
    docker_volumes = __salt__["docker.volumes"]()["Volumes"]
    if docker_volumes:
        volumes = [v for v in docker_volumes if v["Name"] == name]
        if volumes:
            return volumes[0]

    return None


def present(name, driver=None, driver_opts=None, force=False):
    """
    Ensure that a volume is present.

    .. versionadded:: 2015.8.4
    .. versionchanged:: 2015.8.6
        This state no longer deletes and re-creates a volume if the existing
        volume's driver does not match the ``driver`` parameter (unless the
        ``force`` parameter is set to ``True``).
    .. versionchanged:: 2017.7.0
        This state was renamed from **docker.volume_present** to **docker_volume.present**

    name
        Name of the volume

    driver
        Type of driver for that volume.  If ``None`` and the volume
        does not yet exist, the volume will be created using Docker's
        default driver.  If ``None`` and the volume does exist, this
        function does nothing, even if the existing volume's driver is
        not the Docker default driver.  (To ensure that an existing
        volume's driver matches the Docker default, you must
        explicitly name Docker's default driver here.)

    driver_opts
        Options for the volume driver

    force : False
        If the volume already exists but the existing volume's driver
        does not match the driver specified by the ``driver``
        parameter, this parameter controls whether the function errors
        out (if ``False``) or deletes and re-creates the volume (if
        ``True``).

        .. versionadded:: 2015.8.6

    Usage Examples:

    .. code-block:: yaml

        volume_foo:
          docker_volume.present


    .. code-block:: yaml

        volume_bar:
          docker_volume.present
            - name: bar
            - driver: local
            - driver_opts:
                foo: bar

    .. code-block:: yaml

        volume_bar:
          docker_volume.present
            - name: bar
            - driver: local
            - driver_opts:
                - foo: bar
                - option: value

    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    if salt.utils.data.is_dictlist(driver_opts):
        driver_opts = salt.utils.data.repack_dictlist(driver_opts)
    volume = _find_volume(name)
    if not volume:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The volume '{}' will be created".format(name)
            return ret
        try:
            ret["changes"]["created"] = __salt__["docker.create_volume"](
                name, driver=driver, driver_opts=driver_opts
            )
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = "Failed to create volume '{}': {}".format(name, exc)
            return ret
        else:
            result = True
            ret["result"] = result
            return ret
    # volume exists, check if driver is the same.
    if driver is not None and volume["Driver"] != driver:
        if not force:
            ret["comment"] = (
                "Driver for existing volume '{}' ('{}')"
                " does not match specified driver ('{}')"
                " and force is False".format(name, volume["Driver"], driver)
            )
            ret["result"] = None if __opts__["test"] else False
            return ret
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = (
                "The volume '{}' will be replaced with a"
                " new one using the driver '{}'".format(name, volume)
            )
            return ret
        try:
            ret["changes"]["removed"] = __salt__["docker.remove_volume"](name)
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = "Failed to remove volume '{}': {}".format(name, exc)
            return ret
        else:
            try:
                ret["changes"]["created"] = __salt__["docker.create_volume"](
                    name, driver=driver, driver_opts=driver_opts
                )
            except Exception as exc:  # pylint: disable=broad-except
                ret["comment"] = "Failed to create volume '{}': {}".format(name, exc)
                return ret
            else:
                result = True
                ret["result"] = result
                return ret

    ret["result"] = True
    ret["comment"] = "Volume '{}' already exists.".format(name)
    return ret


def absent(name, driver=None):
    """
    Ensure that a volume is absent.

    .. versionadded:: 2015.8.4
    .. versionchanged:: 2017.7.0
        This state was renamed from **docker.volume_absent** to **docker_volume.absent**

    name
        Name of the volume

    Usage Examples:

    .. code-block:: yaml

        volume_foo:
          docker_volume.absent

    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    volume = _find_volume(name)
    if not volume:
        ret["result"] = True
        ret["comment"] = "Volume '{}' already absent".format(name)
        return ret

    try:
        ret["changes"]["removed"] = __salt__["docker.remove_volume"](name)
        ret["result"] = True
    except Exception as exc:  # pylint: disable=broad-except
        ret["comment"] = "Failed to remove volume '{}': {}".format(name, exc)
    return ret



import pytest

import salt.modules.groupadd as groupadd
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

try:
    import grp
except ImportError:
    pass


pytestmark = [
    pytest.mark.skip_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {groupadd: {}}


def test_add():
    """
    Tests if specified group was added
    """
    with patch(
        "salt.utils.path.which",
        MagicMock(side_effect=[None, "/bin/groupadd", "/bin/groupadd"]),
    ):
        with pytest.raises(CommandExecutionError):
            groupadd.add("test", 100)

        mock = MagicMock(return_value={"retcode": 0})
        with patch.dict(groupadd.__salt__, {"cmd.run_all": mock}):
            assert groupadd.add("test", 100) is True

        with patch.dict(groupadd.__grains__, {"kernel": "Linux"}):
            with patch.dict(groupadd.__salt__, {"cmd.run_all": mock}):
                assert groupadd.add("test", 100, True) is True


def test_info():
    """
    Tests the return of group information
    """
    getgrnam = grp.struct_group(("foo", "*", 20, ["test"]))
    with patch("grp.getgrnam", MagicMock(return_value=getgrnam)):
        ret = {"passwd": "*", "gid": 20, "name": "foo", "members": ["test"]}
        assert groupadd.info("foo") == ret


def test_format_info():
    """
    Tests the formatting of returned group information
    """
    group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
    with patch("salt.modules.groupadd._format_info", MagicMock(return_value=group)):
        data = grp.struct_group(("wheel", "*", 0, ["root"]))
        ret = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
        assert groupadd._format_info(data) == ret


def test_getent():
    """
    Tests the return of information on all groups
    """
    getgrnam = grp.struct_group(("foo", "*", 20, ["test"]))
    with patch("grp.getgrall", MagicMock(return_value=[getgrnam])):
        ret = [{"passwd": "*", "gid": 20, "name": "foo", "members": ["test"]}]
        assert groupadd.getent() == ret


def test_chgid_gid_same():
    """
    Tests if the group id is the same as argument
    """
    mock = MagicMock(return_value={"gid": 10})
    with patch.object(groupadd, "info", mock):
        assert groupadd.chgid("test", 10) is True


def test_chgid():
    """
    Tests the gid for a named group was changed
    """
    with patch(
        "salt.utils.path.which",
        MagicMock(side_effect=[None, "/bin/groupmod"]),
    ):
        cmd_mock = MagicMock(return_value=None)
        with patch.dict(groupadd.__salt__, {"cmd.run": cmd_mock}):
            info_mock = MagicMock(side_effect=[{"gid": 10}, {"gid": 10}, {"gid": 500}])
            with patch.object(groupadd, "info", info_mock):
                with pytest.raises(CommandExecutionError):
                    groupadd.chgid("test", 500)
                assert groupadd.chgid("test", 500) is True


def test_delete():
    """
    Tests if the specified group was deleted
    """
    with patch(
        "salt.utils.path.which",
        MagicMock(side_effect=[None, "/bin/groupdel"]),
    ):
        with pytest.raises(CommandExecutionError):
            groupadd.delete("test")

        mock_ret = MagicMock(return_value={"retcode": 0})
        with patch.dict(groupadd.__salt__, {"cmd.run_all": mock_ret}):
            assert groupadd.delete("test") is True


def test_adduser():
    """
    Tests if specified user gets added in the group.
    """
    os_version_list = [
        {
            "grains": {
                "kernel": "Linux",
                "os_family": "RedHat",
                "osmajorrelease": "5",
            },
            "cmd": ["/bin/gpasswd", "-a", "root", "test"],
        },
        {
            "grains": {
                "kernel": "Linux",
                "os_family": "Suse",
                "osmajorrelease": "11",
            },
            "cmd": ["/bin/usermod", "-A", "test", "root"],
        },
        {
            "grains": {"kernel": "Linux"},
            "cmd": ["/bin/gpasswd", "--add", "root", "test"],
        },
        {
            "grains": {"kernel": "OTHERKERNEL"},
            "cmd": ["/bin/usermod", "-G", "test", "root"],
        },
    ]
    with patch(
        "salt.utils.path.which",
        MagicMock(
            side_effect=[
                None,
                "/bin/gpasswd",
                "/bin/usermod",
                "/bin/gpasswd",
                "/bin/usermod",
            ]
        ),
    ):
        with patch.dict(groupadd.__grains__, os_version_list[0]["grains"]):
            with pytest.raises(CommandExecutionError):
                groupadd.adduser("test", "root")

        for os_version in os_version_list:
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(groupadd.__grains__, os_version["grains"]), patch.dict(
                groupadd.__salt__, {"cmd.retcode": mock}
            ):
                assert groupadd.adduser("test", "root") is False
                groupadd.__salt__["cmd.retcode"].assert_called_once_with(
                    os_version["cmd"], python_shell=False
                )


def test_deluser():
    """
    Tests if specified user gets deleted from the group.
    """
    os_version_list = [
        {
            "grains": {
                "kernel": "Linux",
                "os_family": "RedHat",
                "osmajorrelease": "5",
            },
            "cmd": ["/bin/gpasswd", "-d", "root", "test"],
        },
        {
            "grains": {
                "kernel": "Linux",
                "os_family": "Suse",
                "osmajorrelease": "11",
            },
            "cmd": ["/bin/usermod", "-R", "test", "root"],
        },
        {
            "grains": {"kernel": "Linux"},
            "cmd": ["/bin/gpasswd", "--del", "root", "test"],
        },
        {"grains": {"kernel": "OpenBSD"}, "cmd": ["/bin/usermod", "-S", "foo", "root"]},
    ]

    with patch(
        "salt.utils.path.which",
        MagicMock(
            side_effect=[
                None,
                "/bin/gpasswd",
                "/bin/usermod",
                "/bin/gpasswd",
                "/bin/usermod",
            ]
        ),
    ):
        with patch.dict(groupadd.__grains__, os_version_list[0]["grains"]), patch.dict(
            groupadd.__salt__,
            {
                "group.info": MagicMock(
                    return_value={
                        "passwd": "*",
                        "gid": 0,
                        "name": "test",
                        "members": ["root"],
                    }
                ),
            },
        ):
            with pytest.raises(CommandExecutionError):
                groupadd.deluser("test", "root")

        for os_version in os_version_list:
            mock_retcode = MagicMock(return_value=0)
            mock_stdout = MagicMock(return_value="test foo")
            mock_info = MagicMock(
                return_value={
                    "passwd": "*",
                    "gid": 0,
                    "name": "test",
                    "members": ["root"],
                }
            )

            with patch.dict(groupadd.__grains__, os_version["grains"]):
                with patch.dict(
                    groupadd.__salt__,
                    {
                        "cmd.retcode": mock_retcode,
                        "group.info": mock_info,
                        "cmd.run_stdout": mock_stdout,
                    },
                ):
                    assert groupadd.deluser("test", "root") is True
                    groupadd.__salt__["cmd.retcode"].assert_called_once_with(
                        os_version["cmd"], python_shell=False
                    )


def test_members():
    """
    Tests if members of the group, get replaced with a provided list.
    """
    os_version_list = [
        {
            "grains": {
                "kernel": "Linux",
                "os_family": "RedHat",
                "osmajorrelease": "5",
            },
            "cmd": ["/bin/gpasswd", "-M", "foo", "test"],
        },
        {
            "grains": {
                "kernel": "Linux",
                "os_family": "Suse",
                "osmajorrelease": "11",
            },
            "cmd": ["/bin/groupmod", "-A", "foo", "test"],
        },
        {
            "grains": {"kernel": "Linux"},
            "cmd": ["/bin/gpasswd", "--members", "foo", "test"],
        },
        {"grains": {"kernel": "OpenBSD"}, "cmd": ["/bin/usermod", "-G", "test", "foo"]},
    ]

    with patch(
        "salt.utils.path.which",
        MagicMock(
            side_effect=[
                None,
                "/bin/gpasswd",
                "/bin/gpasswd",
                "/bin/groupmod",
                "/bin/gpasswd",
                "/bin/groupdel",
                "/bin/groupadd",
                "/bin/usermod",
            ]
        ),
    ):
        with patch.dict(groupadd.__grains__, os_version_list[0]["grains"]):
            with pytest.raises(CommandExecutionError):
                groupadd.members("test", "foo")

        for os_version in os_version_list:
            mock_ret = MagicMock(return_value={"retcode": 0})
            mock_stdout = MagicMock(return_value={"cmd.run_stdout": 1})
            mock_info = MagicMock(
                return_value={
                    "passwd": "*",
                    "gid": 0,
                    "name": "test",
                    "members": ["root"],
                }
            )
            mock = MagicMock(return_value=True)

            with patch.dict(groupadd.__grains__, os_version["grains"]):
                with patch.dict(
                    groupadd.__salt__,
                    {
                        "cmd.retcode": mock_ret,
                        "group.info": mock_info,
                        "cmd.run_stdout": mock_stdout,
                        "cmd.run": mock,
                    },
                ):
                    assert groupadd.members("test", "foo") is False
                    groupadd.__salt__["cmd.retcode"].assert_called_once_with(
                        os_version["cmd"], python_shell=False
                    )



import pytest

from tests.support.unit import TestCase

try:
    import salt.utils.win_service as win_service
except Exception as exc:  # pylint: disable=broad-except
    win_service = exc


class WinServiceImportTestCase(TestCase):
    def test_import(self):
        """
        Simply importing should not raise an error, especially on Linux
        """
        if isinstance(win_service, Exception):
            raise Exception(
                "Importing win_system caused traceback: {}".format(win_service)
            )


@pytest.mark.skip_unless_on_windows
class WinServiceTestCase(TestCase):
    """
    Test cases for salt.utils.win_service
    """

    def test_info(self):
        """
        Test service.info
        """
        # Get info about the spooler service
        info = win_service.info("spooler")

        # Make sure it returns these fields
        field_names = [
            "BinaryPath",
            "ControlsAccepted",
            "Dependencies",
            "Description",
            "DisplayName",
            "ErrorControl",
            "LoadOrderGroup",
            "ServiceAccount",
            "ServiceType",
            "StartType",
            "StartTypeDelayed",
            "Status",
            "Status_CheckPoint",
            "Status_ExitCode",
            "Status_ServiceCode",
            "Status_WaitHint",
            "TagID",
            "sid",
        ]
        for field_name in field_names:
            self.assertIn(field_name, info)

        # Make sure it returns a valid Display Name
        self.assertEqual(info["DisplayName"], "Print Spooler")



import pytest

from tests.support.case import ModuleCase


@pytest.mark.skip_unless_on_windows
@pytest.mark.windows_whitelisted
class WinServermanagerTest(ModuleCase):
    """
    Test for salt.modules.win_servermanager
    """

    @pytest.mark.slow_test
    def test_list_available(self):
        """
        Test list available features to install
        """
        cmd = self.run_function("win_servermanager.list_available")
        self.assertIn("DNS", cmd)
        self.assertIn("NetworkController", cmd)
        self.assertIn("RemoteAccess", cmd)



"""
Integration tests for the lxd states
"""
import pytest

import salt.modules.lxd
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin


@pytest.mark.skipif(salt.modules.lxd.HAS_PYLXD is False, reason="pylxd not installed")
@pytest.mark.skip_if_binaries_missing("lxd", reason="LXD not installed")
@pytest.mark.skip_if_binaries_missing("lxc", reason="LXC not installed")
class LxdProfileTestCase(ModuleCase, SaltReturnAssertsMixin):
    def tearDown(self):
        self.run_state(
            "lxd_profile.absent",
            name="test-profile",
        )

    def test_02__create_profile(self):
        self.run_state(
            "lxd_profile.absent",
            name="test-profile",
        )
        ret = self.run_state(
            "lxd_profile.present",
            name="test-profile",
            config=[{"key": "boot.autostart", "value": 1}],
        )
        name = "lxd_profile_|-test-profile_|-test-profile_|-present"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]["changes"] == {
            "created": 'Profile "test-profile" has been created'
        }

    def test_03__change_profile(self):
        self.run_state(
            "lxd_profile.present",
            name="test-profile",
            config=[{"key": "boot.autostart", "value": 1}],
        )
        ret = self.run_state(
            "lxd_profile.present",
            name="test-profile",
            config=[
                {"key": "boot.autostart", "value": 1},
                {"key": "security.privileged", "value": "1"},
            ],
        )
        name = "lxd_profile_|-test-profile_|-test-profile_|-present"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]["changes"]["config"] == {
            "security.privileged": 'Added config key "security.privileged" = "1"'
        }

    def test_04__delete_profile(self):
        self.run_state(
            "lxd_profile.present",
            name="test-profile",
            config=[{"key": "boot.autostart", "value": 1}],
        )
        ret = self.run_state(
            "lxd_profile.absent",
            name="test-profile",
        )
        name = "lxd_profile_|-test-profile_|-test-profile_|-absent"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]["changes"] == {
            "removed": 'Profile "test-profile" has been deleted.'
        }



import pytest

import salt.fileserver.svnfs as svnfs
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {svnfs: {}}


def test_env_is_exposed():
    """
    test _env_is_exposed method when
    base is in whitelist
    """
    with patch.dict(
        svnfs.__opts__,
        {"svnfs_saltenv_whitelist": "base", "svnfs_saltenv_blacklist": ""},
    ):
        assert svnfs._env_is_exposed("base")


def test_env_is_exposed_blacklist():
    """
    test _env_is_exposed method when
    base is in blacklist
    """
    with patch.dict(
        svnfs.__opts__,
        {"svnfs_saltenv_whitelist": "", "svnfs_saltenv_blacklist": "base"},
    ):
        assert not svnfs._env_is_exposed("base")



"""
This module exists to help PyInstaller bundle Salt
"""
import pathlib

PYINSTALLER_UTILS_DIR_PATH = pathlib.Path(__file__).resolve().parent


def get_hook_dirs():
    """
    Return a list of paths that PyInstaller can search for hooks.
    """
    hook_dirs = {PYINSTALLER_UTILS_DIR_PATH}
    for path in PYINSTALLER_UTILS_DIR_PATH.iterdir():
        if not path.is_dir():
            continue
        if "__pycache__" in path.parts:
            continue
        hook_dirs.add(path)

    return sorted(str(p) for p in hook_dirs)



"""
Manage groups on Solaris

.. important::
    If you feel that Salt should be using this module to manage groups on a
    minion, and it is using a different module (or gives an error similar to
    *'group.info' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import logging

try:
    import grp
except ImportError:
    pass

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "group"


def __virtual__():
    """
    Set the group module if the kernel is AIX
    """
    if __grains__["kernel"] == "AIX":
        return __virtualname__
    return (
        False,
        "The aix_group execution module failed to load: only available on AIX systems.",
    )


def add(name, gid=None, system=False, root=None, **kwargs):
    """
    Add the specified group

    CLI Example:

    .. code-block:: bash

        salt '*' group.add foo 3456
    """
    cmd = "mkgroup "
    if system and root is not None:
        cmd += "-a "

    if gid:
        cmd += "id={} ".format(gid)

    cmd += name

    ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    return not ret["retcode"]


def delete(name):
    """
    Remove the named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.delete foo
    """
    ret = __salt__["cmd.run_all"]("rmgroup {}".format(name), python_shell=False)

    return not ret["retcode"]


def info(name):
    """
    Return information about a group

    CLI Example:

    .. code-block:: bash

        salt '*' group.info foo
    """
    try:
        grinfo = grp.getgrnam(name)
    except KeyError:
        return {}
    else:
        return {
            "name": grinfo.gr_name,
            "passwd": grinfo.gr_passwd,
            "gid": grinfo.gr_gid,
            "members": grinfo.gr_mem,
        }


def getent(refresh=False):
    """
    Return info on all groups

    CLI Example:

    .. code-block:: bash

        salt '*' group.getent
    """
    if "group.getent" in __context__ and not refresh:
        return __context__["group.getent"]

    ret = []
    for grinfo in grp.getgrall():
        ret.append(info(grinfo.gr_name))

    __context__["group.getent"] = ret
    return ret


def chgid(name, gid):
    """
    Change the gid for a named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.chgid foo 4376
    """
    pre_gid = __salt__["file.group_to_gid"](name)
    if gid == pre_gid:
        return True
    cmd = "chgroup id={} {}".format(gid, name)
    __salt__["cmd.run"](cmd, python_shell=False)
    post_gid = __salt__["file.group_to_gid"](name)
    if post_gid != pre_gid:
        return post_gid == gid
    return False


def adduser(name, username, root=None):
    """
    Add a user in the group.

    CLI Example:

    .. code-block:: bash

         salt '*' group.adduser foo bar

    Verifies if a valid username 'bar' as a member of an existing group 'foo',
    if not then adds it.
    """
    cmd = "chgrpmem -m + {} {}".format(username, name)

    retcode = __salt__["cmd.retcode"](cmd, python_shell=False)

    return not retcode


def deluser(name, username, root=None):
    """
    Remove a user from the group.

    CLI Example:

    .. code-block:: bash

         salt '*' group.deluser foo bar

    Removes a member user 'bar' from a group 'foo'. If group is not present
    then returns True.
    """
    grp_info = __salt__["group.info"](name)
    try:
        if username in grp_info["members"]:
            cmd = "chgrpmem -m - {} {}".format(username, name)
            ret = __salt__["cmd.run"](cmd, python_shell=False)
            return not ret["retcode"]
        else:
            return True
    except Exception:  # pylint: disable=broad-except
        return True


def members(name, members_list, root=None):
    """
    Replaces members of the group with a provided list.

    CLI Example:

    .. code-block:: bash

        salt '*' group.members foo 'user1,user2,user3,...'

    Replaces a membership list for a local group 'foo'.
        foo:x:1234:user1,user2,user3,...
    """
    cmd = "chgrpmem -m = {} {}".format(members_list, name)
    retcode = __salt__["cmd.retcode"](cmd, python_shell=False)

    return not retcode



import logging
import sys
import warnings

import pytest

from salt.utils.functools import namespaced_function
from tests.conftest import CODE_DIR

log = logging.getLogger(__name__)


def preserve_context_ids(value):
    return "preserve_context={}".format(value)


@pytest.fixture(params=[True, False], ids=preserve_context_ids)
def preserve_context(request):
    return request.param


def test_namespacing(tmp_path, shell):
    pkgpath = tmp_path / "foopkg"
    mod1_contents = """
    import json
    import time

    def func_1():
        return time.time()

    def main():
        data = {
            "func1": func_1(),
            "module": func_1.__module__,
            "time_present": "time" in func_1.__globals__
        }
        print(json.dumps(data))

    if __name__ == "__main__":
        main()
    """
    mod2_contents = """
    import json
    from salt.utils.functools import namespaced_function
    from foopkg.mod1 import func_1

    func_1 = namespaced_function(func_1, globals())

    def main():
        data = {
            "func1": func_1(),
            "module": func_1.__module__,
            "time_present": "time" in func_1.__globals__
        }
        print(json.dumps(data))

    if __name__ == "__main__":
        main()
    """
    run1_contents = """
    import sys
    sys.path.insert(0, '{}')
    import foopkg.mod1

    foopkg.mod1.main()
    """.format(
        CODE_DIR
    )
    run2_contents = """
    import sys
    sys.path.insert(0, '{}')
    import foopkg.mod2

    foopkg.mod2.main()
    """.format(
        CODE_DIR
    )
    with pytest.helpers.temp_file(
        "run1.py", contents=run1_contents, directory=tmp_path
    ), pytest.helpers.temp_file(
        "run2.py", contents=run2_contents, directory=tmp_path
    ), pytest.helpers.temp_file(
        "__init__.py", contents="", directory=pkgpath
    ), pytest.helpers.temp_file(
        "mod1.py", mod1_contents, directory=pkgpath
    ), pytest.helpers.temp_file(
        "mod2.py", mod2_contents, directory=pkgpath
    ):
        ret = shell.run(sys.executable, str(tmp_path / "run1.py"), cwd=str(tmp_path))
        log.warning(ret)
        assert ret.returncode == 0
        assert ret.data["module"] == "foopkg.mod1"
        assert ret.data["time_present"] is True
        assert isinstance(ret.data["func1"], float)
        ret = shell.run(sys.executable, str(tmp_path / "run2.py"), cwd=str(tmp_path))
        log.warning(ret)
        assert ret.returncode == 0
        assert ret.data["module"] == "foopkg.mod2"
        assert isinstance(ret.data["func1"], float)
        assert ret.data["time_present"] is True


def test_deprecated_defaults_kwarg():
    def foo():
        pass

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        namespaced_function(foo, globals(), defaults={"foo": 1})

    assert str(w[-1].message) == (
        "Passing 'defaults' to 'namespaced_function' is deprecated, slated "
        "for removal in 3008.0 (Argon) and no longer does anything for the function "
        "being namespaced."
    )


def test_deprecated_preserve_context_kwarg(preserve_context):
    def foo():
        pass

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        namespaced_function(foo, globals(), preserve_context=preserve_context)

    assert str(w[-1].message) == (
        "Passing 'preserve_context' to 'namespaced_function' is deprecated, slated "
        "for removal in 3008.0 (Argon) and no longer does anything for the function "
        "being namespaced."
    )



import logging
import os
import pathlib

import pytest

import salt.modules.win_file as win_file
import salt.modules.win_lgpo as lgpo_mod
import salt.states.win_lgpo as lgpo
import salt.utils.win_dacl as win_dacl

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cachedir = tmp_path / "__test_admx_policy_cache_dir"
    cachedir.mkdir(parents=True, exist_ok=True)
    return {
        lgpo_mod: {
            "__salt__": {
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
            },
            "__opts__": {
                "cachedir": str(cachedir),
            },
            "__context__": {},
        },
        lgpo: {
            "__salt__": {
                "lgpo.get_policy_info": lgpo_mod.get_policy_info,
                "lgpo.get_policy": lgpo_mod.get_policy,
                "lgpo.set": lgpo_mod.set_,
            },
            "__opts__": {
                "test": False,
            },
        },
        win_file: {
            "__utils__": {
                "dacl.set_perms": win_dacl.set_perms,
            },
        },
    }


@pytest.fixture
def clean_comp():
    reg_pol = pathlib.Path(
        os.getenv("SystemRoot"), "System32", "GroupPolicy", "Machine", "Registry.pol"
    )
    reg_pol.unlink(missing_ok=True)
    yield reg_pol
    reg_pol.unlink(missing_ok=True)


def test_allow_telemetry_subsequent_runs(clean_comp):
    """
    Tests that the AllowTelemetry policy is applied correctly and that it
    doesn't appear in subsequent group policy states as having changed
    """
    # Set an initial state for RA_Unsolicit
    result = lgpo_mod.set_computer_policy(name="RA_Unsolicit", setting="Not Configured")
    assert result is True
    # Set AllowTelemetry
    result = lgpo_mod.set_computer_policy(name="AllowTelemetry", setting="Disabled")
    assert result is True
    expected = {
        "new": {"Computer Configuration": {"RA_Unsolicit": "Disabled"}},
        "old": {"Computer Configuration": {"RA_Unsolicit": "Not Configured"}},
    }
    # Set RA_Unsolicit with a state. AllowTelemetry should NOT be in the results
    result = lgpo.set_(
        name="RA_Unsolicit",
        setting="Disabled",
        policy_class="Machine",
    )
    # Run it again and there should be no changes
    assert result["changes"] == expected
    result = lgpo.set_(
        name="RA_Unsolicit",
        setting="Disabled",
        policy_class="Machine",
    )
    assert result["changes"] == {}



"""
    Logstash Logging Handler
    ========================

    .. versionadded:: 0.17.0

    This module provides some `Logstash`_ logging handlers.


    UDP Logging Handler
    -------------------

    For versions of `Logstash`_ before 1.2.0:

    In the salt configuration file:

    .. code-block:: yaml

        logstash_udp_handler:
          host: 127.0.0.1
          port: 9999
          version: 0
          msg_type: logstash

    In the `Logstash`_ configuration file:

    .. code-block:: text

        input {
          udp {
            type => "udp-type"
            format => "json_event"
          }
        }

    For version 1.2.0 of `Logstash`_ and newer:

    In the salt configuration file:

    .. code-block:: yaml

        logstash_udp_handler:
          host: 127.0.0.1
          port: 9999
          version: 1
          msg_type: logstash

    In the `Logstash`_ configuration file:

    .. code-block:: text

        input {
          udp {
            port => 9999
            codec => json
          }
        }

    Please read the `UDP input`_ configuration page for additional information.


    ZeroMQ Logging Handler
    ----------------------

    For versions of `Logstash`_ before 1.2.0:

    In the salt configuration file:

    .. code-block:: yaml

        logstash_zmq_handler:
          address: tcp://127.0.0.1:2021
          version: 0

    In the `Logstash`_ configuration file:

    .. code-block:: text

        input {
          zeromq {
            type => "zeromq-type"
            mode => "server"
            topology => "pubsub"
            address => "tcp://0.0.0.0:2021"
            charset => "UTF-8"
            format => "json_event"
          }
        }

    For version 1.2.0 of `Logstash`_ and newer:

    In the salt configuration file:

    .. code-block:: yaml

        logstash_zmq_handler:
          address: tcp://127.0.0.1:2021
          version: 1

    In the `Logstash`_ configuration file:

    .. code-block:: text

        input {
          zeromq {
            topology => "pubsub"
            address => "tcp://0.0.0.0:2021"
            codec => json
          }
        }

    Please read the `ZeroMQ input`_ configuration page for additional
    information.

    .. admonition:: Important Logstash Setting

        One of the most important settings that you should not forget on your
        `Logstash`_ configuration file regarding these logging handlers is
        ``format``.
        Both the `UDP` and `ZeroMQ` inputs need to have ``format`` as
        ``json_event`` which is what we send over the wire.


    Log Level
    .........

    Both the ``logstash_udp_handler`` and the ``logstash_zmq_handler``
    configuration sections accept an additional setting ``log_level``. If not
    set, the logging level used will be the one defined for ``log_level`` in
    the global configuration file section.

    HWM
    ...

    The `high water mark`_ for the ZMQ socket setting. Only applicable for the
    ``logstash_zmq_handler``.



    .. admonition:: Inspiration

        This work was inspired in `pylogstash`_, `python-logstash`_, `canary`_
        and the `PyZMQ logging handler`_.


    .. _`Logstash`: http://logstash.net
    .. _`canary`: https://github.com/ryanpetrello/canary
    .. _`pylogstash`: https://github.com/turtlebender/pylogstash
    .. _`python-logstash`: https://github.com/vklochan/python-logstash
    .. _`PyZMQ logging handler`: https://github.com/zeromq/pyzmq/blob/master/zmq/log/handlers.py
    .. _`UDP input`: http://logstash.net/docs/latest/inputs/udp
    .. _`ZeroMQ input`: http://logstash.net/docs/latest/inputs/zeromq
    .. _`high water mark`: http://api.zeromq.org/3-2:zmq-setsockopt

"""


import datetime
import logging
import logging.handlers
import os

import salt.utils.json
import salt.utils.network
import salt.utils.stringutils
from salt._logging import LOG_LEVELS

try:
    import zmq
    import zmq.error
except ImportError:
    pass

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "logstash"


def __virtual__():
    if not any(
        ["logstash_udp_handler" in __opts__, "logstash_zmq_handler" in __opts__]
    ):
        log.trace(
            "None of the required configuration sections, "
            "'logstash_udp_handler' and 'logstash_zmq_handler', "
            "were found in the configuration. Not loading the Logstash "
            "logging handlers module."
        )
        return False
    return __virtualname__


def setup_handlers():
    host = port = address = None

    if "logstash_udp_handler" in __opts__:
        host = __opts__["logstash_udp_handler"].get("host", None)
        port = __opts__["logstash_udp_handler"].get("port", None)
        version = __opts__["logstash_udp_handler"].get("version", 0)
        msg_type = __opts__["logstash_udp_handler"].get("msg_type", "logstash")

        if host is None and port is None:
            log.debug(
                "The required 'logstash_udp_handler' configuration keys, "
                "'host' and/or 'port', are not properly configured. Not "
                "configuring the logstash UDP logging handler."
            )
        else:
            logstash_formatter = LogstashFormatter(msg_type=msg_type, version=version)
            udp_handler = DatagramLogstashHandler(host, port)
            udp_handler.setFormatter(logstash_formatter)
            udp_handler.setLevel(
                LOG_LEVELS[
                    __opts__["logstash_udp_handler"].get(
                        "log_level",
                        # Not set? Get the main salt log_level setting on the
                        # configuration file
                        __opts__.get(
                            "log_level",
                            # Also not set?! Default to 'error'
                            "error",
                        ),
                    )
                ]
            )
            yield udp_handler

    if "logstash_zmq_handler" in __opts__:
        address = __opts__["logstash_zmq_handler"].get("address", None)
        zmq_hwm = __opts__["logstash_zmq_handler"].get("hwm", 1000)
        version = __opts__["logstash_zmq_handler"].get("version", 0)

        if address is None:
            log.debug(
                "The required 'logstash_zmq_handler' configuration key, "
                "'address', is not properly configured. Not "
                "configuring the logstash ZMQ logging handler."
            )
        else:
            logstash_formatter = LogstashFormatter(version=version)
            zmq_handler = ZMQLogstashHander(address, zmq_hwm=zmq_hwm)
            zmq_handler.setFormatter(logstash_formatter)
            zmq_handler.setLevel(
                LOG_LEVELS[
                    __opts__["logstash_zmq_handler"].get(
                        "log_level",
                        # Not set? Get the main salt log_level setting on the
                        # configuration file
                        __opts__.get(
                            "log_level",
                            # Also not set?! Default to 'error'
                            "error",
                        ),
                    )
                ]
            )
            yield zmq_handler

    if host is None and port is None and address is None:
        yield False


class LogstashFormatter(logging.Formatter):
    def __init__(self, msg_type="logstash", msg_path="logstash", version=0):
        self.msg_path = msg_path
        self.msg_type = msg_type
        self.version = version
        self.format = getattr(self, "format_v{}".format(version))
        super().__init__(fmt=None, datefmt=None)

    def formatTime(self, record, datefmt=None):
        return datetime.datetime.utcfromtimestamp(record.created).isoformat()[:-3] + "Z"

    def format_v0(self, record):
        host = salt.utils.network.get_fqhostname()
        message_dict = {
            "@timestamp": self.formatTime(record),
            "@fields": {
                "levelname": record.levelname,
                "logger": record.name,
                "lineno": record.lineno,
                "pathname": record.pathname,
                "process": record.process,
                "threadName": record.threadName,
                "funcName": record.funcName,
                "processName": record.processName,
            },
            "@message": record.getMessage(),
            "@source": "{}://{}/{}".format(self.msg_type, host, self.msg_path),
            "@source_host": host,
            "@source_path": self.msg_path,
            "@tags": ["salt"],
            "@type": self.msg_type,
        }

        if record.exc_info:
            message_dict["@fields"]["exc_info"] = self.formatException(record.exc_info)

        # Add any extra attributes to the message field
        for key, value in record.__dict__.items():
            if key in (
                "args",
                "asctime",
                "created",
                "exc_info",
                "exc_text",
                "filename",
                "funcName",
                "id",
                "levelname",
                "levelno",
                "lineno",
                "module",
                "msecs",
                "msecs",
                "message",
                "msg",
                "name",
                "pathname",
                "process",
                "processName",
                "relativeCreated",
                "thread",
                "threadName",
            ):
                # These are already handled above or not handled at all
                continue

            if value is None:
                message_dict["@fields"][key] = value
                continue

            if isinstance(value, (str, bool, dict, float, int, list)):
                message_dict["@fields"][key] = value
                continue

            message_dict["@fields"][key] = repr(value)
        return salt.utils.json.dumps(message_dict)

    def format_v1(self, record):
        message_dict = {
            "@version": 1,
            "@timestamp": self.formatTime(record),
            "host": salt.utils.network.get_fqhostname(),
            "levelname": record.levelname,
            "logger": record.name,
            "lineno": record.lineno,
            "pathname": record.pathname,
            "process": record.process,
            "threadName": record.threadName,
            "funcName": record.funcName,
            "processName": record.processName,
            "message": record.getMessage(),
            "tags": ["salt"],
            "type": self.msg_type,
        }

        if record.exc_info:
            message_dict["exc_info"] = self.formatException(record.exc_info)

        # Add any extra attributes to the message field
        for key, value in record.__dict__.items():
            if key in (
                "args",
                "asctime",
                "created",
                "exc_info",
                "exc_text",
                "filename",
                "funcName",
                "id",
                "levelname",
                "levelno",
                "lineno",
                "module",
                "msecs",
                "msecs",
                "message",
                "msg",
                "name",
                "pathname",
                "process",
                "processName",
                "relativeCreated",
                "thread",
                "threadName",
            ):
                # These are already handled above or not handled at all
                continue

            if value is None:
                message_dict[key] = value
                continue

            if isinstance(value, (str, bool, dict, float, int, list)):
                message_dict[key] = value
                continue

            message_dict[key] = repr(value)
        return salt.utils.json.dumps(message_dict)


class DatagramLogstashHandler(logging.handlers.DatagramHandler):
    """
    Logstash UDP logging handler.
    """

    def makePickle(self, record):
        return salt.utils.stringutils.to_bytes(self.format(record))


class ZMQLogstashHander(logging.Handler):
    """
    Logstash ZMQ logging handler.
    """

    def __init__(self, address, level=logging.NOTSET, zmq_hwm=1000):
        super().__init__(level=level)
        self._context = self._publisher = None
        self._address = address
        self._zmq_hwm = zmq_hwm
        self._pid = os.getpid()

    @property
    def publisher(self):
        current_pid = os.getpid()
        if not getattr(self, "_publisher") or self._pid != current_pid:
            # We forked? Multiprocessing? Recreate!!!
            self._pid = current_pid
            self._context = zmq.Context()
            self._publisher = self._context.socket(zmq.PUB)
            # Above 1000 unsent events in the socket queue, stop dropping them
            try:
                # Above the defined high water mark(unsent messages), start
                # dropping them
                self._publisher.setsockopt(zmq.HWM, self._zmq_hwm)
            except (AttributeError, zmq.error.ZMQError):
                # In ZMQ >= 3.0, there are separate send and receive HWM
                # settings
                self._publisher.setsockopt(zmq.SNDHWM, self._zmq_hwm)
                self._publisher.setsockopt(zmq.RCVHWM, self._zmq_hwm)

            self._publisher.connect(self._address)
        return self._publisher

    def emit(self, record):
        formatted_object = salt.utils.stringutils.to_bytes(self.format(record))
        self.publisher.send(formatted_object)

    def close(self):
        if self._context is not None:
            # One second to send any queued messages
            if hasattr(self._context, "destroy"):
                self._context.destroy(1 * 1000)
            else:
                if getattr(self, "_publisher", None) is not None:
                    self._publisher.setsockopt(zmq.LINGER, 1 * 1000)
                    self._publisher.close()

                if self._context.closed is False:
                    self._context.term()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.npm as npmmod
import salt.states.npm as npm
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, create_autospec, patch


@pytest.fixture
def configure_loader_modules():
    return {npm: {"__opts__": {"test": False}}}


@pytest.fixture(params=["", {}, []])
def fake_install(request):
    fake_install = create_autospec(npmmod.install, return_value=request.param)
    with patch.dict(
        npm.__salt__,
        {
            "npm.list": create_autospec(npmmod.list_, return_value={}),
            "npm.install": fake_install,
        },
    ):
        yield fake_install


def test_when_install_does_not_error_installed_should_be_true(fake_install):
    ret = npm.installed("fnord")
    assert ret["result"] is True


def test_installed():
    """
    Test to verify that the given package is installed
    and is at the correct version.
    """
    name = "coffee-script"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    mock_err = MagicMock(side_effect=CommandExecutionError)
    mock_dict = MagicMock(return_value={name: {"version": "1.2"}})
    with patch.dict(npm.__salt__, {"npm.list": mock_err}):
        comt = "Error looking up 'coffee-script': "
        ret.update({"comment": comt})
        assert npm.installed(name) == ret

    with patch.dict(npm.__salt__, {"npm.list": mock_dict, "npm.install": mock_err}):
        with patch.dict(npm.__opts__, {"test": True}):
            comt = "Package(s) 'coffee-script' satisfied by coffee-script@1.2"
            ret.update({"comment": comt, "result": True})
            assert npm.installed(name) == ret

        with patch.dict(npm.__opts__, {"test": False}):
            comt = "Package(s) 'coffee-script' satisfied by coffee-script@1.2"
            ret.update({"comment": comt, "result": True})
            assert npm.installed(name) == ret

            comt = "Error installing 'n, p, m': "
            ret.update({"comment": comt, "result": False})
            assert npm.installed(name, "npm") == ret

            with patch.dict(npm.__salt__, {"npm.install": mock_dict}):
                comt = "Package(s) 'n, p, m' successfully installed"
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "changes": {"new": ["n", "p", "m"], "old": []},
                    }
                )
                assert npm.installed(name, "npm") == ret


def test_removed():
    """
    Test to verify that the given package is not installed.
    """
    name = "coffee-script"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    mock_err = MagicMock(
        side_effect=[CommandExecutionError, {}, {name: ""}, {name: ""}]
    )
    mock_t = MagicMock(return_value=True)
    with patch.dict(npm.__salt__, {"npm.list": mock_err, "npm.uninstall": mock_t}):
        comt = "Error uninstalling 'coffee-script': "
        ret.update({"comment": comt})
        assert npm.removed(name) == ret

        comt = "Package 'coffee-script' is not installed"
        ret.update({"comment": comt, "result": True})
        assert npm.removed(name) == ret

        with patch.dict(npm.__opts__, {"test": True}):
            comt = "Package 'coffee-script' is set to be removed"
            ret.update({"comment": comt, "result": None})
            assert npm.removed(name) == ret

        with patch.dict(npm.__opts__, {"test": False}):
            comt = "Package 'coffee-script' was successfully removed"
            ret.update({"comment": comt, "result": True, "changes": {name: "Removed"}})
            assert npm.removed(name) == ret


def test_bootstrap():
    """
    Test to bootstraps a node.js application.
    """
    name = "coffee-script"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    mock_err = MagicMock(side_effect=[CommandExecutionError, False, True])
    with patch.dict(npm.__salt__, {"npm.install": mock_err}):
        comt = "Error Bootstrapping 'coffee-script': "
        ret.update({"comment": comt})
        assert npm.bootstrap(name) == ret

        comt = "Directory is already bootstrapped"
        ret.update({"comment": comt, "result": True})
        assert npm.bootstrap(name) == ret

        comt = "Directory was successfully bootstrapped"
        ret.update({"comment": comt, "result": True, "changes": {name: "Bootstrapped"}})
        assert npm.bootstrap(name) == ret


def test_cache_cleaned():
    """
    Test to verify that the npm cache is cleaned.
    """
    name = "coffee-script"

    pkg_ret = {"name": name, "result": False, "comment": "", "changes": {}}
    ret = {"name": None, "result": False, "comment": "", "changes": {}}

    mock_list = MagicMock(return_value=["~/.npm", "~/.npm/{}/".format(name)])
    mock_cache_clean_success = MagicMock(return_value=True)
    mock_cache_clean_failure = MagicMock(return_value=False)
    mock_err = MagicMock(side_effect=CommandExecutionError)

    with patch.dict(npm.__salt__, {"npm.cache_list": mock_err}):
        comt = "Error looking up cached packages: "
        ret.update({"comment": comt})
        assert npm.cache_cleaned() == ret

    with patch.dict(npm.__salt__, {"npm.cache_list": mock_err}):
        comt = "Error looking up cached {}: ".format(name)
        pkg_ret.update({"comment": comt})
        assert npm.cache_cleaned(name) == pkg_ret

    mock_data = {"npm.cache_list": mock_list, "npm.cache_clean": MagicMock()}
    with patch.dict(npm.__salt__, mock_data):
        non_cached_pkg = "salt"
        comt = "Package {} is not in the cache".format(non_cached_pkg)
        pkg_ret.update({"name": non_cached_pkg, "result": True, "comment": comt})
        assert npm.cache_cleaned(non_cached_pkg) == pkg_ret
        pkg_ret.update({"name": name})

        with patch.dict(npm.__opts__, {"test": True}):
            comt = "Cached packages set to be removed"
            ret.update({"result": None, "comment": comt})
            assert npm.cache_cleaned() == ret

        with patch.dict(npm.__opts__, {"test": True}):
            comt = "Cached {} set to be removed".format(name)
            pkg_ret.update({"result": None, "comment": comt})
            assert npm.cache_cleaned(name) == pkg_ret

        with patch.dict(npm.__opts__, {"test": False}):
            comt = "Cached packages successfully removed"
            ret.update(
                {"result": True, "comment": comt, "changes": {"cache": "Removed"}}
            )
            assert npm.cache_cleaned() == ret

        with patch.dict(npm.__opts__, {"test": False}):
            comt = "Cached {} successfully removed".format(name)
            pkg_ret.update(
                {"result": True, "comment": comt, "changes": {name: "Removed"}}
            )
            assert npm.cache_cleaned(name) == pkg_ret

    mock_data = {
        "npm.cache_list": mock_list,
        "npm.cache_clean": MagicMock(return_value=False),
    }
    with patch.dict(npm.__salt__, mock_data):
        with patch.dict(npm.__opts__, {"test": False}):
            comt = "Error cleaning cached packages"
            ret.update({"result": False, "comment": comt})
            ret["changes"] = {}
            assert npm.cache_cleaned() == ret

        with patch.dict(npm.__opts__, {"test": False}):
            comt = "Error cleaning cached {}".format(name)
            pkg_ret.update({"result": False, "comment": comt})
            pkg_ret["changes"] = {}
            assert npm.cache_cleaned(name) == pkg_ret



"""
Support for htpasswd module. Requires the apache2-utils package for Debian-based distros.

.. versionadded:: 2014.7.0

.. code-block:: yaml

    username:
      webutil.user_exists:
        - password: secr3t
        - htpasswd_file: /etc/nginx/htpasswd
        - options: d
        - force: true

"""


import salt.utils.path

__virtualname__ = "webutil"


def __virtual__():
    """
    depends on webutil module
    """
    if salt.utils.path.which("htpasswd"):
        return __virtualname__
    return (False, "Command not found: htpasswd")


def user_exists(
    name,
    password=None,
    htpasswd_file=None,
    options="",
    force=False,
    runas=None,
    update=False,
):
    """
    Make sure the user is inside the specified htpasswd file

    name
        User name

    password
        User password

    htpasswd_file
        Path to the htpasswd file

    options
        See :mod:`salt.modules.htpasswd.useradd`

    force
        Touch the file even if user already created

    runas
        The system user to run htpasswd command with

    update
        Update an existing user's password if it's different from what's in
        the htpasswd file (unlike force, which updates regardless)

    """
    ret = {"name": name, "changes": {}, "comment": "", "result": None}

    if __salt__["file.file_exists"](htpasswd_file):
        exists = (
            __salt__["file.grep"](htpasswd_file, "^{}:".format(name))["retcode"] == 0
        )
    else:
        exists = False

    # If user exists, but we're supposed to update the password, find out if
    # it's changed, but not if we're forced to update the file regardless.
    password_changed = False
    if exists and update and not force:
        password_changed = not __salt__["webutil.verify"](
            htpasswd_file, name, password, opts=options, runas=runas
        )

    if not exists or password_changed or force:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "User '{}' is set to be added to htpasswd file".format(
                name
            )
            ret["changes"] = {name: True}
            return ret

        useradd_ret = __salt__["webutil.useradd"](
            htpasswd_file, name, password, opts=options, runas=runas
        )
        if useradd_ret["retcode"] == 0:
            ret["result"] = True
            ret["comment"] = useradd_ret["stderr"]
            ret["changes"] = {name: True}
            return ret
        else:
            ret["result"] = False
            ret["comment"] = useradd_ret["stderr"]
            return ret

    if __opts__["test"] and ret["changes"]:
        ret["result"] = None
    else:
        ret["result"] = True
    ret["comment"] = "User already known"
    return ret


def user_absent(name, htpasswd_file=None, runas=None):
    """
    Make sure the user is not in the specified htpasswd file

    name
        User name

    htpasswd_file
        Path to the htpasswd file

    runas
        The system user to run htpasswd command with

    """
    ret = {"name": name, "changes": {}, "comment": "", "result": None}

    exists = __salt__["file.grep"](htpasswd_file, "^{}:".format(name))["retcode"] == 0

    if not exists:
        if __opts__["test"]:
            ret["result"] = None
        else:
            ret["result"] = True
        ret["comment"] = "User already not in file"
    else:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "User '{}' is set to be removed from htpasswd file".format(
                name
            )
            ret["changes"] = {name: True}
        else:
            userdel_ret = __salt__["webutil.userdel"](
                htpasswd_file, name, runas=runas, all_results=True
            )

            ret["result"] = userdel_ret["retcode"] == 0
            ret["comment"] = userdel_ret["stderr"]

            if ret["result"]:
                ret["changes"] = {name: True}

    return ret



import pytest

import salt.states.virt as virt
from tests.support.mock import MagicMock, patch

from .helpers import network_update_call


@pytest.fixture
def configure_loader_modules(libvirt_mock):
    return {virt: {"libvirt": libvirt_mock}}


def test_network_defined_not_existing(test):
    """
    network_defined state tests if the network doesn't exist yet.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        define_mock = MagicMock(return_value=True)
        with patch.dict(
            virt.__salt__,
            {
                "virt.network_info": MagicMock(
                    side_effect=[{}, {"mynet": {"active": False}}]
                ),
                "virt.network_define": define_mock,
            },
        ):
            assert virt.network_defined(
                "mynet",
                "br2",
                "bridge",
                vport="openvswitch",
                tag=180,
                ipv4_config={
                    "cidr": "192.168.2.0/24",
                    "dhcp_ranges": [
                        {"start": "192.168.2.10", "end": "192.168.2.25"},
                        {"start": "192.168.2.110", "end": "192.168.2.125"},
                    ],
                },
                ipv6_config={
                    "cidr": "2001:db8:ca2:2::1/64",
                    "dhcp_ranges": [
                        {"start": "2001:db8:ca2:1::10", "end": "2001:db8:ca2::1f"},
                    ],
                },
                mtu=9000,
                domain={"name": "acme.lab"},
                nat={"ports": {"start": 1024, "end": 2048}},
                interfaces="eth0 eth1",
                addresses="0000:01:02.4 0000:01:02.5",
                physical_function="eth4",
                dns={
                    "hosts": {
                        "192.168.2.10": {"name": "web", "mac": "de:ad:be:ef:00:00"}
                    }
                },
                autostart=False,
                connection="myconnection",
                username="user",
                password="secret",
            ) == {
                "name": "mynet",
                "changes": {"mynet": "Network defined"},
                "result": None if test else True,
                "comment": "Network mynet defined",
            }
            if not test:
                define_mock.assert_called_with(
                    "mynet",
                    "br2",
                    "bridge",
                    vport="openvswitch",
                    tag=180,
                    autostart=False,
                    start=False,
                    ipv4_config={
                        "cidr": "192.168.2.0/24",
                        "dhcp_ranges": [
                            {"start": "192.168.2.10", "end": "192.168.2.25"},
                            {"start": "192.168.2.110", "end": "192.168.2.125"},
                        ],
                    },
                    ipv6_config={
                        "cidr": "2001:db8:ca2:2::1/64",
                        "dhcp_ranges": [
                            {"start": "2001:db8:ca2:1::10", "end": "2001:db8:ca2::1f"},
                        ],
                    },
                    mtu=9000,
                    domain={"name": "acme.lab"},
                    nat={"ports": {"start": 1024, "end": 2048}},
                    interfaces="eth0 eth1",
                    addresses="0000:01:02.4 0000:01:02.5",
                    physical_function="eth4",
                    dns={
                        "hosts": {
                            "192.168.2.10": {"name": "web", "mac": "de:ad:be:ef:00:00"}
                        }
                    },
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
            else:
                define_mock.assert_not_called()


def test_network_defined_no_change(test):
    """
    network_defined state tests if the network doesn't need update.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        define_mock = MagicMock(return_value=True)
        update_mock = MagicMock(return_value=False)
        with patch.dict(
            virt.__salt__,
            {
                "virt.network_info": MagicMock(
                    return_value={"mynet": {"active": True, "autostart": True}}
                ),
                "virt.network_define": define_mock,
                "virt.network_update": update_mock,
            },
        ):
            assert virt.network_defined("mynet", "br2", "bridge") == {
                "name": "mynet",
                "changes": {},
                "result": True,
                "comment": "Network mynet unchanged",
            }
            define_mock.assert_not_called()
            assert update_mock.call_args_list == [
                network_update_call("mynet", "br2", "bridge", test=True)
            ]


def test_network_defined_change(test):
    """
    network_defined state tests if the network needs update.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        define_mock = MagicMock(return_value=True)
        update_mock = MagicMock(return_value=True)
        autostart_mock = MagicMock(return_value=True)
        with patch.dict(
            virt.__salt__,
            {
                "virt.network_info": MagicMock(
                    return_value={"mynet": {"active": True, "autostart": True}}
                ),
                "virt.network_define": define_mock,
                "virt.network_update": update_mock,
                "virt.network_set_autostart": autostart_mock,
            },
        ):
            assert virt.network_defined(
                "mynet",
                "br2",
                "bridge",
                vport="openvswitch",
                tag=180,
                ipv4_config={
                    "cidr": "192.168.2.0/24",
                    "dhcp_ranges": [
                        {"start": "192.168.2.10", "end": "192.168.2.25"},
                        {"start": "192.168.2.110", "end": "192.168.2.125"},
                    ],
                },
                ipv6_config={
                    "cidr": "2001:db8:ca2:2::1/64",
                    "dhcp_ranges": [
                        {"start": "2001:db8:ca2:1::10", "end": "2001:db8:ca2::1f"},
                    ],
                },
                mtu=9000,
                domain={"name": "acme.lab"},
                nat={"ports": {"start": 1024, "end": 2048}},
                interfaces="eth0 eth1",
                addresses="0000:01:02.4 0000:01:02.5",
                physical_function="eth4",
                dns={
                    "hosts": {
                        "192.168.2.10": {"name": "web", "mac": "de:ad:be:ef:00:00"}
                    }
                },
                autostart=False,
                connection="myconnection",
                username="user",
                password="secret",
            ) == {
                "name": "mynet",
                "changes": {"mynet": "Network updated, autostart flag changed"},
                "result": None if test else True,
                "comment": "Network mynet updated, autostart flag changed",
            }
            define_mock.assert_not_called()
            expected_update_kwargs = {
                "vport": "openvswitch",
                "tag": 180,
                "ipv4_config": {
                    "cidr": "192.168.2.0/24",
                    "dhcp_ranges": [
                        {"start": "192.168.2.10", "end": "192.168.2.25"},
                        {"start": "192.168.2.110", "end": "192.168.2.125"},
                    ],
                },
                "ipv6_config": {
                    "cidr": "2001:db8:ca2:2::1/64",
                    "dhcp_ranges": [
                        {"start": "2001:db8:ca2:1::10", "end": "2001:db8:ca2::1f"},
                    ],
                },
                "mtu": 9000,
                "domain": {"name": "acme.lab"},
                "nat": {"ports": {"start": 1024, "end": 2048}},
                "interfaces": "eth0 eth1",
                "addresses": "0000:01:02.4 0000:01:02.5",
                "physical_function": "eth4",
                "dns": {
                    "hosts": {
                        "192.168.2.10": {"name": "web", "mac": "de:ad:be:ef:00:00"}
                    }
                },
                "connection": "myconnection",
                "username": "user",
                "password": "secret",
            }
            calls = [
                network_update_call(
                    "mynet", "br2", "bridge", **expected_update_kwargs, test=True
                )
            ]
            if test:
                assert update_mock.call_args_list == calls
                autostart_mock.assert_not_called()
            else:
                calls.append(
                    network_update_call(
                        "mynet", "br2", "bridge", **expected_update_kwargs, test=False
                    )
                )
                assert update_mock.call_args_list == calls
                autostart_mock.assert_called_with(
                    "mynet",
                    state="off",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )


def test_network_defined_error(test):
    """
    network_defined state tests if an error is triggered by libvirt.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        define_mock = MagicMock(return_value=True)
        with patch.dict(
            virt.__salt__,
            {
                "virt.network_info": MagicMock(
                    side_effect=virt.libvirt.libvirtError("Some error")
                )
            },
        ):
            assert virt.network_defined("mynet", "br2", "bridge") == {
                "name": "mynet",
                "changes": {},
                "result": False,
                "comment": "Some error",
            }
            define_mock.assert_not_called()


def test_network_running_not_existing(test):
    """
    network_running state test cases, non-existing network case.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        define_mock = MagicMock(return_value=True)
        start_mock = MagicMock(return_value=True)
        # Non-existing network case
        with patch.dict(
            virt.__salt__,
            {
                "virt.network_info": MagicMock(
                    side_effect=[{}, {"mynet": {"active": False}}]
                ),
                "virt.network_define": define_mock,
                "virt.network_start": start_mock,
            },
        ):
            assert virt.network_running(
                "mynet",
                "br2",
                "bridge",
                vport="openvswitch",
                tag=180,
                ipv4_config={
                    "cidr": "192.168.2.0/24",
                    "dhcp_ranges": [
                        {"start": "192.168.2.10", "end": "192.168.2.25"},
                        {"start": "192.168.2.110", "end": "192.168.2.125"},
                    ],
                },
                ipv6_config={
                    "cidr": "2001:db8:ca2:2::1/64",
                    "dhcp_ranges": [
                        {"start": "2001:db8:ca2:1::10", "end": "2001:db8:ca2::1f"},
                    ],
                },
                mtu=9000,
                domain={"name": "acme.lab"},
                nat={"ports": {"start": 1024, "end": 2048}},
                interfaces="eth0 eth1",
                addresses="0000:01:02.4 0000:01:02.5",
                physical_function="eth4",
                dns={
                    "hosts": {
                        "192.168.2.10": {"name": "web", "mac": "de:ad:be:ef:00:00"}
                    }
                },
                autostart=False,
                connection="myconnection",
                username="user",
                password="secret",
            ) == {
                "name": "mynet",
                "changes": {"mynet": "Network defined and started"},
                "comment": "Network mynet defined and started",
                "result": None if test else True,
            }
            if not test:
                define_mock.assert_called_with(
                    "mynet",
                    "br2",
                    "bridge",
                    vport="openvswitch",
                    tag=180,
                    autostart=False,
                    start=False,
                    ipv4_config={
                        "cidr": "192.168.2.0/24",
                        "dhcp_ranges": [
                            {"start": "192.168.2.10", "end": "192.168.2.25"},
                            {"start": "192.168.2.110", "end": "192.168.2.125"},
                        ],
                    },
                    ipv6_config={
                        "cidr": "2001:db8:ca2:2::1/64",
                        "dhcp_ranges": [
                            {"start": "2001:db8:ca2:1::10", "end": "2001:db8:ca2::1f"},
                        ],
                    },
                    mtu=9000,
                    domain={"name": "acme.lab"},
                    nat={"ports": {"start": 1024, "end": 2048}},
                    interfaces="eth0 eth1",
                    addresses="0000:01:02.4 0000:01:02.5",
                    physical_function="eth4",
                    dns={
                        "hosts": {
                            "192.168.2.10": {"name": "web", "mac": "de:ad:be:ef:00:00"}
                        }
                    },
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
                start_mock.assert_called_with(
                    "mynet",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
            else:
                define_mock.assert_not_called()
                start_mock.assert_not_called()


def test_network_running_nochange(test):
    """
    network_running state test cases, no change case case.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        define_mock = MagicMock(return_value=True)
        update_mock = MagicMock(return_value=False)
        with patch.dict(
            virt.__salt__,
            {
                "virt.network_info": MagicMock(
                    return_value={"mynet": {"active": True, "autostart": True}}
                ),
                "virt.network_define": define_mock,
                "virt.network_update": update_mock,
            },
        ):
            assert virt.network_running("mynet", "br2", "bridge") == {
                "name": "mynet",
                "changes": {},
                "comment": "Network mynet unchanged and is running",
                "result": None if test else True,
            }
            assert update_mock.call_args_list == [
                network_update_call("mynet", "br2", "bridge", test=True)
            ]


def test_network_running_stopped(test):
    """
    network_running state test cases, network stopped case.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        define_mock = MagicMock(return_value=True)
        start_mock = MagicMock(return_value=True)
        update_mock = MagicMock(return_value=False)
        with patch.dict(
            virt.__salt__,
            {  # pylint: disable=no-member
                "virt.network_info": MagicMock(
                    return_value={"mynet": {"active": False, "autostart": True}}
                ),
                "virt.network_start": start_mock,
                "virt.network_define": define_mock,
                "virt.network_update": update_mock,
            },
        ):
            assert virt.network_running(
                "mynet",
                "br2",
                "bridge",
                connection="myconnection",
                username="user",
                password="secret",
            ) == {
                "name": "mynet",
                "changes": {"mynet": "Network started"},
                "comment": "Network mynet unchanged and started",
                "result": None if test else True,
            }
            assert update_mock.call_args_list == [
                network_update_call(
                    "mynet",
                    "br2",
                    "bridge",
                    connection="myconnection",
                    username="user",
                    password="secret",
                    test=True,
                )
            ]
            if not test:
                start_mock.assert_called_with(
                    "mynet",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
            else:
                start_mock.assert_not_called()


def test_network_running_error(test):
    """
    network_running state test cases, libvirt error case.
    """
    with patch.dict(virt.__opts__, {"test": test}):
        with patch.dict(
            virt.__salt__,
            {
                "virt.network_info": MagicMock(
                    side_effect=virt.libvirt.libvirtError("Some error")
                ),
            },
        ):
            assert virt.network_running("mynet", "br2", "bridge") == {
                "name": "mynet",
                "changes": {},
                "comment": "Some error",
                "result": False,
            }



"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""

import builtins
import io

import salt.modules.cron as cron
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, mock_open, patch
from tests.support.unit import TestCase

STUB_USER = "root"
STUB_PATH = "/tmp"

STUB_CRON_TIMESTAMP = {
    "minute": "1",
    "hour": "2",
    "daymonth": "3",
    "month": "4",
    "dayweek": "5",
}

STUB_SIMPLE_RAW_CRON = "5 0 * * * /tmp/no_script.sh"
STUB_SIMPLE_CRON_DICT = {
    "pre": ["5 0 * * * /tmp/no_script.sh"],
    "crons": [],
    "env": [],
    "special": [],
}
STUB_CRON_SPACES = """
# Lines below here are managed by Salt, do not edit
TEST_VAR="a string with plenty of spaces"
# SALT_CRON_IDENTIFIER:echo "must  be  double  spaced"
11 * * * * echo "must  be  double  spaced"
"""
STUB_AT_SIGN = """
# Lines below here are managed by Salt, do not edit
# SALT_CRON_IDENTIFIER:echo "cron with @ sign"
@daily echo "cron with @ sign"
@daily
"""
STUB_NO_AT_SIGN = """
# Lines below here are managed by Salt, do not edit
# SALT_CRON_IDENTIFIER:echo "cron without @ sign"
1 2 3 4 5 echo "cron without @ sign"
"""

L = "# Lines below here are managed by Salt, do not edit\n"

CRONTAB = io.StringIO()


def get_crontab(*args, **kw):
    return CRONTAB.getvalue()


def set_crontab(val):
    CRONTAB.seek(0)
    CRONTAB.truncate(0)
    CRONTAB.write(val)


def write_crontab(*args, **kw):
    set_crontab("\n".join([a.strip() for a in args[1]]))
    return MagicMock()


class CronTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {cron: {}}

    def test__need_changes_new(self):
        """
        New behavior, identifier will get track of the managed lines!
        """
        with patch(
            "salt.modules.cron.raw_cron", new=MagicMock(side_effect=get_crontab)
        ), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(side_effect=write_crontab),
        ):
            # when there are no identifiers,
            # we do not touch it
            set_crontab(L + "# SALT_CRON_IDENTIFIER:booh\n* * * * * ls\n")
            cron.set_job(
                user="root",
                minute="*",
                hour="*",
                daymonth="*",
                month="*",
                dayweek="*",
                cmd="ls",
                comment=None,
                identifier=None,
            )
            c1 = get_crontab()
            set_crontab(L + "* * * * * ls\n")
            self.assertEqual(
                c1,
                "# Lines below here are managed by Salt, do not edit\n"
                "# SALT_CRON_IDENTIFIER:booh\n"
                "* * * * * ls\n"
                "* * * * * ls",
            )
            # whenever we have an identifier, hourray even without comment
            # we can match and edit the crontab in place
            # without cluttering the crontab with new cmds
            set_crontab(L + "# SALT_CRON_IDENTIFIER:bar\n* * * * * ls\n")
            cron.set_job(
                user="root",
                minute="*",
                hour="*",
                daymonth="*",
                month="*",
                dayweek="*",
                cmd="ls",
                comment=None,
                identifier="bar",
            )
            c5 = get_crontab()
            set_crontab(L + "* * * * * ls\n")
            self.assertEqual(
                c5,
                "# Lines below here are managed by Salt, do not edit\n"
                "# SALT_CRON_IDENTIFIER:bar\n"
                "* * * * * ls\n",
            )
            # we can even change the other parameters as well
            # thx to the id
            set_crontab(L + "# SALT_CRON_IDENTIFIER:bar\n* * * * * ls\n")
            cron.set_job(
                user="root",
                minute="1",
                hour="2",
                daymonth="3",
                month="4",
                dayweek="5",
                cmd="foo",
                comment="moo",
                identifier="bar",
            )
            c6 = get_crontab()
            self.assertEqual(
                c6,
                "# Lines below here are managed by Salt, do not edit\n"
                "# moo SALT_CRON_IDENTIFIER:bar\n"
                "1 2 3 4 5 foo",
            )

    def test__unicode_match(self):
        with patch.object(builtins, "__salt_system_encoding__", "utf-8"):
            self.assertTrue(cron._cron_matched({"identifier": "1"}, "foo", 1))
            self.assertTrue(cron._cron_matched({"identifier": "é"}, "foo", "é"))
            self.assertTrue(cron._cron_matched({"identifier": "é"}, "foo", "é"))
            self.assertTrue(cron._cron_matched({"identifier": "é"}, "foo", "é"))
            self.assertTrue(cron._cron_matched({"identifier": "é"}, "foo", "é"))

    def test__need_changes_old(self):
        """
        old behavior; ID has no special action
        - If an id is found, it will be added as a new crontab
          even if there is a cmd that looks like this one
        - no comment, delete the cmd and readd it
        - comment: idem
        """
        with patch(
            "salt.modules.cron.raw_cron", new=MagicMock(side_effect=get_crontab)
        ), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(side_effect=write_crontab),
        ):
            set_crontab(L + "* * * * * ls\n\n")
            cron.set_job(
                user="root",
                minute="*",
                hour="*",
                daymonth="*",
                month="*",
                dayweek="*",
                cmd="ls",
                comment=None,
                identifier=cron.SALT_CRON_NO_IDENTIFIER,
            )
            c1 = get_crontab()
            set_crontab(L + "* * * * * ls\n")
            self.assertEqual(
                c1,
                "# Lines below here are managed by Salt, do not edit\n* * * * * ls\n\n",
            )
            cron.set_job(
                user="root",
                minute="*",
                hour="*",
                daymonth="*",
                month="*",
                dayweek="*",
                cmd="ls",
                comment="foo",
                identifier=cron.SALT_CRON_NO_IDENTIFIER,
            )
            c2 = get_crontab()
            self.assertEqual(
                c2,
                "# Lines below here are managed by Salt, do not edit\n"
                "# foo\n* * * * * ls",
            )
            set_crontab(L + "* * * * * ls\n")
            cron.set_job(
                user="root",
                minute="*",
                hour="*",
                daymonth="*",
                month="*",
                dayweek="*",
                cmd="lsa",
                comment="foo",
                identifier="bar",
            )
            c3 = get_crontab()
            self.assertEqual(
                c3,
                "# Lines below here are managed by Salt, do not edit\n"
                "* * * * * ls\n"
                "# foo SALT_CRON_IDENTIFIER:bar\n"
                "* * * * * lsa",
            )
            set_crontab(L + "* * * * * ls\n")
            cron.set_job(
                user="root",
                minute="*",
                hour="*",
                daymonth="*",
                month="*",
                dayweek="*",
                cmd="foo",
                comment="foo",
                identifier="bar",
            )
            c4 = get_crontab()
            self.assertEqual(
                c4,
                "# Lines below here are managed by Salt, do not edit\n"
                "* * * * * ls\n"
                "# foo SALT_CRON_IDENTIFIER:bar\n"
                "* * * * * foo",
            )
            set_crontab(L + "* * * * * ls\n")
            cron.set_job(
                user="root",
                minute="*",
                hour="*",
                daymonth="*",
                month="*",
                dayweek="*",
                cmd="ls",
                comment="foo",
                identifier="bbar",
            )
            c4 = get_crontab()
            self.assertEqual(
                c4,
                "# Lines below here are managed by Salt, do not edit\n"
                "# foo SALT_CRON_IDENTIFIER:bbar\n"
                "* * * * * ls",
            )

    def test__issue10959(self):
        """
        handle multi old style crontabs
        https://github.com/saltstack/salt/issues/10959
        """
        with patch(
            "salt.modules.cron.raw_cron", new=MagicMock(side_effect=get_crontab)
        ), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(side_effect=write_crontab),
        ):
            set_crontab(
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                # as managed per salt, the last lines will be merged together !
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * samecmd\n"
                "* * * * * otheridcmd\n"
                "* * * * * otheridcmd\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n0 * * * * samecmd1\n"
                "1 * * * * samecmd1\n"
                "0 * * * * otheridcmd1\n"
                "1 * * * * otheridcmd1\n"
                # special case here, none id managed line with same command
                # as a later id managed line will become managed
                "# SALT_CRON_IDENTIFIER:1\n0 * * * * otheridcmd1\n"
                "# SALT_CRON_IDENTIFIER:2\n0 * * * * otheridcmd1\n"
            )
            crons1 = cron.list_tab("root")
            # the filtering is done on save, we reflect in listing
            # the same that we have in a file, no matter what we
            # have
            self.assertEqual(
                crons1,
                {
                    "crons": [
                        {
                            "cmd": "ls",
                            "comment": "uoo",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "NO ID SET",
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "too",
                            "comment": "uuoo",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "NO ID SET",
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "zoo",
                            "comment": "uuuoo",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "NO ID SET",
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "yoo",
                            "comment": "",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "NO ID SET",
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "xoo",
                            "comment": "",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "NO ID SET",
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "samecmd",
                            "comment": "",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "NO ID SET",
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "samecmd",
                            "comment": None,
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": None,
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "otheridcmd",
                            "comment": None,
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": None,
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "otheridcmd",
                            "comment": None,
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": None,
                            "minute": "*",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "samecmd1",
                            "comment": "",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "NO ID SET",
                            "minute": "0",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "samecmd1",
                            "comment": None,
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": None,
                            "minute": "1",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "otheridcmd1",
                            "comment": None,
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": None,
                            "minute": "0",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "otheridcmd1",
                            "comment": None,
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": None,
                            "minute": "1",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "otheridcmd1",
                            "comment": "",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "1",
                            "minute": "0",
                            "month": "*",
                            "commented": False,
                        },
                        {
                            "cmd": "otheridcmd1",
                            "comment": "",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "2",
                            "minute": "0",
                            "month": "*",
                            "commented": False,
                        },
                    ],
                    "env": [],
                    "pre": [],
                    "special": [],
                },
            )
            # so yood so far, no problem for now, trying to save the
            # multilines without id crons now
            inc_tests = [
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n"
                "0 * * * * samecmd1",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n1 * * * * samecmd1",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n1 * * * * samecmd1\n"
                "0 * * * * otheridcmd1",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n1 * * * * samecmd1\n"
                "1 * * * * otheridcmd1",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n1 * * * * samecmd1\n"
                "# SALT_CRON_IDENTIFIER:1\n0 * * * * otheridcmd1",
                #
                "# Lines below here are managed by Salt, do not edit\n"
                "# uoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * ls\n"
                "# uuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * too\n"
                "# uuuoo SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * zoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * yoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * xoo\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n* * * * * samecmd\n"
                "* * * * * otheridcmd\n"
                "# SALT_CRON_IDENTIFIER:NO ID SET\n1 * * * * samecmd1\n"
                "# SALT_CRON_IDENTIFIER:1\n0 * * * * otheridcmd1\n"
                "# SALT_CRON_IDENTIFIER:2\n0 * * * * otheridcmd1",
            ]
            set_crontab("")
            for idx, cr in enumerate(crons1["crons"]):
                cron.set_job("root", **cr)
                self.assertEqual(
                    get_crontab(),
                    inc_tests[idx],
                    ("idx {0}\n'{1}'\n != \n'{2}'\n\n\n'{1}' != '{2}'").format(
                        idx, get_crontab(), inc_tests[idx]
                    ),
                )

    def test_list_tab_commented_cron_jobs(self):
        """
        handle commented cron jobs
        https://github.com/saltstack/salt/issues/29082
        """
        with patch(
            "salt.modules.cron.raw_cron", MagicMock(side_effect=get_crontab)
        ), patch(
            "salt.modules.cron._write_cron_lines", MagicMock(side_effect=write_crontab)
        ):
            set_crontab(
                "# An unmanaged commented cron job\n"
                "#0 * * * * /bin/true\n"
                "# Lines below here are managed by Salt, do not edit\n"
                "# cron_1 SALT_CRON_IDENTIFIER:cron_1\n#DISABLED#0 * * * * my_cmd_1\n"
                "# cron_2 SALT_CRON_IDENTIFIER:cron_2\n#DISABLED#* * * * * my_cmd_2\n"
                "# cron_3 SALT_CRON_IDENTIFIER:cron_3\n"
                "#DISABLED#but it is a comment line"
                "#DISABLED#0 * * * * my_cmd_3\n"
                "# cron_4 SALT_CRON_IDENTIFIER:cron_4\n0 * * * * my_cmd_4\n"
            )
            crons1 = cron.list_tab("root")
            self.assertEqual(
                crons1,
                {
                    "crons": [
                        {
                            "cmd": "my_cmd_1",
                            "comment": "cron_1",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "cron_1",
                            "minute": "0",
                            "month": "*",
                            "commented": True,
                        },
                        {
                            "cmd": "my_cmd_2",
                            "comment": "cron_2",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "cron_2",
                            "minute": "*",
                            "month": "*",
                            "commented": True,
                        },
                        {
                            "cmd": "line#DISABLED#0 * * * * my_cmd_3",
                            "comment": "cron_3",
                            "daymonth": "is",
                            "dayweek": "comment",
                            "hour": "it",
                            "identifier": "cron_3",
                            "minute": "but",
                            "month": "a",
                            "commented": True,
                        },
                        {
                            "cmd": "my_cmd_4",
                            "comment": "cron_4",
                            "daymonth": "*",
                            "dayweek": "*",
                            "hour": "*",
                            "identifier": "cron_4",
                            "minute": "0",
                            "month": "*",
                            "commented": False,
                        },
                    ],
                    "env": [],
                    "pre": [
                        "# An unmanaged commented cron job",
                        "#0 * * * * /bin/true",
                    ],
                    "special": [],
                },
            )

    def test_get_entry(self):
        """
        test get_entry function
        """
        list_tab_output = {
            "crons": [
                {
                    "cmd": "my_cmd_1",
                    "comment": "cron_1",
                    "daymonth": "*",
                    "dayweek": "*",
                    "hour": "*",
                    "identifier": "cron_1",
                    "minute": "0",
                    "month": "*",
                    "commented": True,
                },
                {
                    "cmd": "my_cmd_2",
                    "comment": "cron_2",
                    "daymonth": "*",
                    "dayweek": "*",
                    "hour": "*",
                    "identifier": "cron_2",
                    "minute": "*",
                    "month": "*",
                    "commented": True,
                },
                {
                    "cmd": "line#DISABLED#0 * * * * my_cmd_3",
                    "comment": "cron_3",
                    "daymonth": "is",
                    "dayweek": "comment",
                    "hour": "it",
                    "identifier": "cron_3",
                    "minute": "but",
                    "month": "a",
                    "commented": True,
                },
                {
                    "cmd": "my_cmd_4",
                    "comment": "cron_4",
                    "daymonth": "*",
                    "dayweek": "*",
                    "hour": "*",
                    "identifier": "cron_4",
                    "minute": "0",
                    "month": "*",
                    "commented": False,
                },
            ],
            "env": [],
            "pre": ["# An unmanaged commented cron job", "#0 * * * * /bin/true"],
            "special": [],
        }
        get_entry_2 = {
            "comment": "cron_2",
            "cmd": "my_cmd_2",
            "identifier": "cron_2",
            "dayweek": "*",
            "daymonth": "*",
            "hour": "*",
            "minute": "*",
            "month": "*",
            "commented": True,
        }
        get_entry_3 = {
            "comment": "cron_3",
            "identifier": "cron_3",
            "dayweek": "comment",
            "hour": "it",
            "cmd": "line#DISABLED#0 * * * * my_cmd_3",
            "daymonth": "is",
            "commented": True,
            "minute": "but",
            "month": "a",
        }
        with patch(
            "salt.modules.cron.list_tab", new=MagicMock(return_value=list_tab_output)
        ):
            # Test get_entry identifier
            get_entry_output = cron.get_entry("root", identifier="cron_3")
            self.assertDictEqual(get_entry_output, get_entry_3)
            # Test get_entry cmd
            get_entry_output = cron.get_entry("root", cmd="my_cmd_2")
            self.assertDictEqual(get_entry_output, get_entry_2)
            # Test identifier wins when both specified
            get_entry_output = cron.get_entry(
                "root", identifier="cron_3", cmd="my_cmd_2"
            )
            self.assertDictEqual(get_entry_output, get_entry_3)

    def test_cron_extra_spaces(self):
        """
        Issue #38449
        """
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron.raw_cron", MagicMock(return_value=STUB_CRON_SPACES)
        ):
            ret = cron.list_tab("root")
            eret = {
                "crons": [
                    {
                        "cmd": 'echo "must  be  double  spaced"',
                        "comment": "",
                        "commented": False,
                        "daymonth": "*",
                        "dayweek": "*",
                        "hour": "*",
                        "identifier": 'echo "must  be  double  spaced"',
                        "minute": "11",
                        "month": "*",
                    }
                ],
                "env": [
                    {"name": "TEST_VAR", "value": '"a string with plenty of spaces"'}
                ],
                "pre": [""],
                "special": [],
            }
            self.assertEqual(eret, ret)

    def test_cron_at_sign(self):
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron.raw_cron", MagicMock(return_value=STUB_AT_SIGN)
        ):
            ret = cron.list_tab("root")
            eret = {
                "crons": [],
                "env": [],
                "pre": [""],
                "special": [
                    {
                        "cmd": 'echo "cron with @ sign"',
                        "comment": "",
                        "commented": False,
                        "identifier": 'echo "cron with @ sign"',
                        "spec": "@daily",
                    }
                ],
            }
            self.assertDictEqual(eret, ret)

    def test__load_tab(self):
        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch(
            "salt.modules.cron.raw_cron",
            new=MagicMock(
                side_effect=[
                    (L + "\n"),
                    (L + "* * * * * ls\nn"),
                    (L + "# commented\n#DISABLED#* * * * * ls\n"),
                    (L + "# foo\n* * * * * ls\n"),
                    (
                        L
                        + "# foo {}:blah\n".format(cron.SALT_CRON_IDENTIFIER)
                        + "* * * * * ls\n"
                    ),
                ]
            ),
        ):
            crons1 = cron.list_tab("root")
            crons2 = cron.list_tab("root")
            crons3 = cron.list_tab("root")
            crons4 = cron.list_tab("root")
            crons5 = cron.list_tab("root")
            self.assertEqual(crons1, {"pre": [], "crons": [], "env": [], "special": []})
            self.assertEqual(
                crons2["crons"][0],
                {
                    "comment": None,
                    "commented": False,
                    "dayweek": "*",
                    "hour": "*",
                    "identifier": None,
                    "cmd": "ls",
                    "daymonth": "*",
                    "minute": "*",
                    "month": "*",
                },
            )
            self.assertEqual(
                crons3["crons"][0],
                {
                    "comment": "commented",
                    "commented": True,
                    "dayweek": "*",
                    "hour": "*",
                    "identifier": None,
                    "cmd": "ls",
                    "daymonth": "*",
                    "minute": "*",
                    "month": "*",
                },
            )
            self.assertEqual(
                crons4["crons"][0],
                {
                    "comment": "foo",
                    "commented": False,
                    "dayweek": "*",
                    "hour": "*",
                    "identifier": None,
                    "cmd": "ls",
                    "daymonth": "*",
                    "minute": "*",
                    "month": "*",
                },
            )
            self.assertEqual(
                crons5["crons"][0],
                {
                    "comment": "foo",
                    "commented": False,
                    "dayweek": "*",
                    "hour": "*",
                    "identifier": "blah",
                    "cmd": "ls",
                    "daymonth": "*",
                    "minute": "*",
                    "month": "*",
                },
            )

    def test_write_cron_file_root_rh(self):
        """
        Assert that write_cron_file() is called with the correct cron command and user: RedHat
          - If instance running uid matches crontab user uid, run without -u flag.
        """
        with patch.dict(cron.__grains__, {"os_family": "RedHat"}), patch.dict(
            cron.__salt__, {"cmd.retcode": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match",
            new=MagicMock(return_value=True),
        ):
            cron.write_cron_file(STUB_USER, STUB_PATH)
            cron.__salt__["cmd.retcode"].assert_called_with(
                "crontab /tmp", python_shell=False
            )

    def test_write_cron_file_foo_rh(self):
        """
        Assert that write_cron_file() is called with the correct cron command and user: RedHat
          - If instance running with uid that doesn't match crontab user uid, runas foo
        """
        with patch.dict(cron.__grains__, {"os_family": "RedHat"}), patch.dict(
            cron.__salt__, {"cmd.retcode": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=False)
        ):
            cron.write_cron_file("foo", STUB_PATH)
            cron.__salt__["cmd.retcode"].assert_called_with(
                "crontab /tmp", runas="foo", python_shell=False
            )

    def test_write_cron_file_root_sol(self):
        """
        Assert that write_cron_file() is called with the correct cron command and user: Solaris
          - Solaris should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch.dict(
            cron.__salt__, {"cmd.retcode": MagicMock()}
        ):
            cron.write_cron_file(STUB_USER, STUB_PATH)
            cron.__salt__["cmd.retcode"].assert_called_with(
                "crontab /tmp", runas=STUB_USER, python_shell=False
            )

    def test_write_cron_file_foo_sol(self):
        """
        Assert that write_cron_file() is called with the correct cron command and user: Solaris
          - Solaris should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch.dict(
            cron.__salt__, {"cmd.retcode": MagicMock()}
        ):
            cron.write_cron_file("foo", STUB_PATH)
            cron.__salt__["cmd.retcode"].assert_called_with(
                "crontab /tmp", runas="foo", python_shell=False
            )

    def test_write_cron_file_root_aix(self):
        """
        Assert that write_cron_file() is called with the correct cron command and user: AIX
          - AIX should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "AIX"}), patch.dict(
            cron.__salt__, {"cmd.retcode": MagicMock()}
        ):
            cron.write_cron_file(STUB_USER, STUB_PATH)
            cron.__salt__["cmd.retcode"].assert_called_with(
                "crontab /tmp", runas=STUB_USER, python_shell=False
            )

    def test_write_cron_file_foo_aix(self):
        """
        Assert that write_cron_file() is called with the correct cron command and user: AIX
          - AIX should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "AIX"}), patch.dict(
            cron.__salt__, {"cmd.retcode": MagicMock()}
        ):
            cron.write_cron_file("foo", STUB_PATH)
            cron.__salt__["cmd.retcode"].assert_called_with(
                "crontab /tmp", runas="foo", python_shell=False
            )

    def test_write_cr_file_v_root_rh(self):
        """
        Assert that write_cron_file_verbose() is called with the correct cron command and user: RedHat
          - If instance running uid matches crontab user uid, run without -u flag.
        """
        with patch.dict(cron.__grains__, {"os_family": "Redhat"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=True)
        ):
            cron.write_cron_file_verbose(STUB_USER, STUB_PATH)
            cron.__salt__["cmd.run_all"].assert_called_with(
                "crontab /tmp", python_shell=False
            )

    def test_write_cr_file_v_foo_rh(self):
        """
        Assert that write_cron_file_verbose() is called with the correct cron command and user: RedHat
          - If instance running with uid that doesn't match crontab user uid, runas 'foo'
        """
        with patch.dict(cron.__grains__, {"os_family": "Redhat"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=False)
        ):
            cron.write_cron_file_verbose("foo", STUB_PATH)
            cron.__salt__["cmd.run_all"].assert_called_with(
                "crontab /tmp", runas="foo", python_shell=False
            )

    def test_write_cr_file_v_root_sol(self):
        """
        Assert that write_cron_file_verbose() is called with the correct cron command and user: Solaris
          - Solaris should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ):
            cron.write_cron_file_verbose(STUB_USER, STUB_PATH)
            cron.__salt__["cmd.run_all"].assert_called_with(
                "crontab /tmp", runas=STUB_USER, python_shell=False
            )

    def test_write_cr_file_v_foo_sol(self):
        """
        Assert that write_cron_file_verbose() is called with the correct cron command and user: Solaris
          - Solaris should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ):
            cron.write_cron_file_verbose("foo", STUB_PATH)
            cron.__salt__["cmd.run_all"].assert_called_with(
                "crontab /tmp", runas="foo", python_shell=False
            )

    def test_write_cr_file_v_root_aix(self):
        """
        Assert that write_cron_file_verbose() is called with the correct cron command and user: AIX
          - AIX should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "AIX"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ):
            cron.write_cron_file_verbose(STUB_USER, STUB_PATH)
            cron.__salt__["cmd.run_all"].assert_called_with(
                "crontab /tmp", runas=STUB_USER, python_shell=False
            )

    def test_write_cr_file_v_foo_aix(self):
        """
        Assert that write_cron_file_verbose() is called with the correct cron command and user: AIX
          - AIX should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "AIX"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ):
            cron.write_cron_file_verbose("foo", STUB_PATH)
            cron.__salt__["cmd.run_all"].assert_called_with(
                "crontab /tmp", runas="foo", python_shell=False
            )

    def test_raw_cron_root_redhat(self):
        """
        Assert that raw_cron() is called with the correct cron command and user: RedHat
          - If instance running uid matches crontab user uid, runas STUB_USER without -u flag.
        """
        with patch.dict(cron.__grains__, {"os_family": "Redhat"}), patch.dict(
            cron.__salt__, {"cmd.run_stdout": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=True)
        ):
            cron.raw_cron(STUB_USER)
            cron.__salt__["cmd.run_stdout"].assert_called_with(
                "crontab -l", ignore_retcode=True, rstrip=False, python_shell=False
            )

    def test_raw_cron_foo_redhat(self):
        """
        Assert that raw_cron() is called with the correct cron command and user: RedHat
          - If instance running with uid that doesn't match crontab user uid, run with -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "Redhat"}), patch.dict(
            cron.__salt__, {"cmd.run_stdout": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=False)
        ):
            cron.raw_cron(STUB_USER)
            cron.__salt__["cmd.run_stdout"].assert_called_with(
                "crontab -l",
                runas=STUB_USER,
                ignore_retcode=True,
                rstrip=False,
                python_shell=False,
            )

    def test_raw_cron_root_solaris(self):
        """
        Assert that raw_cron() is called with the correct cron command and user: Solaris
          - Solaris should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch.dict(
            cron.__salt__, {"cmd.run_stdout": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=True)
        ):
            cron.raw_cron(STUB_USER)
            cron.__salt__["cmd.run_stdout"].assert_called_with(
                "crontab -l",
                runas=STUB_USER,
                ignore_retcode=True,
                rstrip=False,
                python_shell=False,
            )

    def test_raw_cron_foo_solaris(self):
        """
        Assert that raw_cron() is called with the correct cron command and user: Solaris
          - Solaris should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch.dict(
            cron.__salt__, {"cmd.run_stdout": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=False)
        ):
            cron.raw_cron(STUB_USER)
            cron.__salt__["cmd.run_stdout"].assert_called_with(
                "crontab -l",
                runas=STUB_USER,
                ignore_retcode=True,
                rstrip=False,
                python_shell=False,
            )

    def test_raw_cron_root_aix(self):
        """
        Assert that raw_cron() is called with the correct cron command and user: AIX
          - AIX should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "AIX"}), patch.dict(
            cron.__salt__, {"cmd.run_stdout": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=True)
        ):
            cron.raw_cron(STUB_USER)
            cron.__salt__["cmd.run_stdout"].assert_called_with(
                "crontab -l",
                runas=STUB_USER,
                ignore_retcode=True,
                rstrip=False,
                python_shell=False,
            )

    def test_raw_cron_foo_aix(self):
        """
        Assert that raw_cron() is called with the correct cron command and user: AIX
          - AIX should always run without a -u flag
        """
        with patch.dict(cron.__grains__, {"os_family": "AIX"}), patch.dict(
            cron.__salt__, {"cmd.run_stdout": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match", MagicMock(return_value=False)
        ):
            cron.raw_cron(STUB_USER)
            cron.__salt__["cmd.run_stdout"].assert_called_with(
                "crontab -l",
                runas=STUB_USER,
                ignore_retcode=True,
                rstrip=False,
                python_shell=False,
            )


class PsTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {cron: {}}

    def test__needs_change(self):
        self.assertTrue(cron._needs_change(True, False))

    def test__needs_change_random(self):
        """
        Assert that if the new var is 'random' and old is '* that we return True
        """
        self.assertTrue(cron._needs_change("*", "random"))

    ## Still trying to figure this one out.
    # def test__render_tab(self):
    #     pass
    def test__get_cron_cmdstr(self):
        self.assertEqual("crontab /tmp", cron._get_cron_cmdstr(STUB_PATH))

    # Test get_cron_cmdstr() when user is added
    def test__get_cron_cmdstr_user(self):
        """
        Passes if a user is added to crontab command
        """
        self.assertEqual(
            "crontab -u root /tmp", cron._get_cron_cmdstr(STUB_PATH, STUB_USER)
        )

    def test__date_time_match(self):
        """
        Passes if a match is found on all elements. Note the conversions to strings here!
        :return:
        """
        self.assertTrue(
            cron._date_time_match(
                STUB_CRON_TIMESTAMP,
                minute=STUB_CRON_TIMESTAMP["minute"],
                hour=STUB_CRON_TIMESTAMP["hour"],
                daymonth=STUB_CRON_TIMESTAMP["daymonth"],
                dayweek=STUB_CRON_TIMESTAMP["dayweek"],
            )
        )

    def test_list_tab(self):
        with patch(
            "salt.modules.cron.raw_cron",
            new=MagicMock(return_value=STUB_SIMPLE_RAW_CRON),
        ):
            self.assertDictEqual(STUB_SIMPLE_CRON_DICT, cron.list_tab("DUMMY_USER"))

    def test_set_special(self):
        with patch(
            "salt.modules.cron._write_cron_lines"
        ) as write_cron_lines_mock, patch(
            "salt.modules.cron.list_tab",
            new=MagicMock(return_value=STUB_SIMPLE_CRON_DICT),
        ):
            expected_write_call = call(
                "DUMMY_USER",
                [
                    "5 0 * * * /tmp/no_script.sh\n",
                    "# Lines below here are managed by Salt, do not edit\n",
                    "@hourly echo Hi!\n",
                ],
            )
            ret = cron.set_special("DUMMY_USER", "@hourly", "echo Hi!")
            write_cron_lines_mock.assert_has_calls(
                (expected_write_call,), any_order=True
            )

    def test_set_special_from_job(self):
        """Use set_special to turn a non-special entry into a special one"""
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(return_value={"retcode": False}),
        ), patch(
            "salt.modules.cron.raw_cron",
            new=MagicMock(side_effect=[STUB_NO_AT_SIGN, STUB_NO_AT_SIGN, L]),
        ):
            expected_call = call(
                "DUMMY_USER",
                [
                    "# Lines below here are managed by Salt, do not edit\n",
                    '# SALT_CRON_IDENTIFIER:echo "cron without @ sign"\n',
                    '@daily echo "cron without @ sign"\n',
                ],
            )
            ret = cron.set_special(
                "DUMMY_USER",
                "@daily",
                'echo "cron without @ sign"',
                identifier='echo "cron without @ sign"',
            )
            cron._write_cron_lines.assert_has_calls((expected_call,), any_order=True)

    def test__get_cron_date_time(self):
        ret = cron._get_cron_date_time(
            minute=STUB_CRON_TIMESTAMP["minute"],
            hour=STUB_CRON_TIMESTAMP["hour"],
            daymonth=STUB_CRON_TIMESTAMP["daymonth"],
            dayweek=STUB_CRON_TIMESTAMP["dayweek"],
            month=STUB_CRON_TIMESTAMP["month"],
        )
        self.assertDictEqual(ret, STUB_CRON_TIMESTAMP)

    def test__get_cron_date_time_daymonth_max(self):
        ret = cron._get_cron_date_time(
            minute="random",
            hour="random",
            daymonth="random",
            dayweek="random",
            month="random",
        )
        self.assertTrue(int(ret["minute"]) in range(0, 60))
        self.assertTrue(int(ret["hour"]) in range(0, 24))
        self.assertTrue(int(ret["daymonth"]) in range(1, 32))
        self.assertTrue(int(ret["dayweek"]) in range(0, 8))
        self.assertTrue(int(ret["month"]) in range(1, 13))

    def test_set_job(self):
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(return_value={"retcode": False}),
        ), patch(
            "salt.modules.cron.raw_cron",
            new=MagicMock(return_value=STUB_SIMPLE_RAW_CRON),
        ):
            cron.set_job(
                "DUMMY_USER",
                1,
                2,
                3,
                4,
                5,
                "/bin/echo NOT A DROID",
                comment="WERE YOU LOOKING FOR ME?",
            )
            expected_call = call(
                "DUMMY_USER",
                [
                    "5 0 * * * /tmp/no_script.sh\n",
                    "# Lines below here are managed by Salt, do not edit\n",
                    "# WERE YOU LOOKING FOR ME?\n",
                    "1 2 3 4 5 /bin/echo NOT A DROID\n",
                ],
            )
            cron._write_cron_lines.assert_has_calls((expected_call,), any_order=True)

    def test_set_job_from_special(self):
        """Use set_job to turn a special entry into a non-special one"""
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(return_value={"retcode": False}),
        ), patch(
            "salt.modules.cron.raw_cron",
            new=MagicMock(side_effect=[STUB_AT_SIGN, STUB_AT_SIGN, L]),
        ):
            cron.set_job(
                "DUMMY_USER",
                1,
                2,
                3,
                4,
                5,
                'echo "cron with @ sign"',
                identifier='echo "cron with @ sign"',
            )
            expected_call = call(
                "DUMMY_USER",
                [
                    "# Lines below here are managed by Salt, do not edit\n",
                    '# SALT_CRON_IDENTIFIER:echo "cron with @ sign"\n',
                    '1 2 3 4 5 echo "cron with @ sign"\n',
                ],
            )
            cron._write_cron_lines.assert_has_calls((expected_call,), any_order=True)

    def test_rm_special(self):
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(return_value={"retcode": False}),
        ), patch(
            "salt.modules.cron.raw_cron", new=MagicMock(return_value=STUB_AT_SIGN)
        ):
            ret = cron.rm_special(
                "root",
                'echo "cron with @ sign"',
                special="@daily",
                identifier='echo "cron with @ sign"',
            )
            self.assertEqual("removed", ret)

    def test_rm_special_default_special(self):
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(return_value={"retcode": False}),
        ), patch(
            "salt.modules.cron.raw_cron", new=MagicMock(return_value=STUB_AT_SIGN)
        ):
            ret = cron.rm_special(
                "root", 'echo "cron with @ sign"', identifier='echo "cron with @ sign"'
            )
            self.assertEqual("removed", ret)

    def test_rm_special_absent(self):
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(return_value={"retcode": False}),
        ), patch(
            "salt.modules.cron.raw_cron", new=MagicMock(return_value=STUB_AT_SIGN)
        ):
            ret = cron.rm_special(
                "root", 'echo "there is no job"', identifier='echo "there is no job"'
            )
            self.assertEqual("absent", ret)

    def test_rm_job_is_absent(self):
        with patch.dict(cron.__grains__, {"os": None}), patch(
            "salt.modules.cron._write_cron_lines",
            new=MagicMock(return_value={"retcode": False}),
        ), patch(
            "salt.modules.cron.raw_cron",
            new=MagicMock(return_value=STUB_SIMPLE_RAW_CRON),
        ):
            ret = cron.rm_job("DUMMY_USER", "/bin/echo NOT A DROID", 1, 2, 3, 4, 5)
            self.assertEqual("absent", ret)

    def test_write_cron_lines_euid_match_user_rh(self):
        """
        Assert that _write_cron_lines() is called with the correct cron command and user
        OS: RedHat. EUID match User (either root, root or user, user).
        Expected to run without runas argument.
        """
        temp_path = "some_temp_path"
        crontab_cmd = "crontab {}".format(temp_path)

        with patch.dict(cron.__grains__, {"os_family": "RedHat"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match",
            new=MagicMock(return_value=True),
        ), patch(
            "salt.utils.files.fpopen", mock_open()
        ), patch(
            "salt.utils.files.mkstemp", MagicMock(return_value=temp_path)
        ), patch(
            "os.remove", MagicMock()
        ):
            cron._write_cron_lines("root", "test 123")
            cron.__salt__["cmd.run_all"].assert_called_with(
                crontab_cmd, python_shell=False
            )

    def test_write_cron_lines_root_non_root_rh(self):
        """
        Assert that _write_cron_lines() is called with the correct cron command and user
        OS: RedHat. EUID: root. User: non-root.
        Expected to run without runas argument and with -u non-root argument.
        """
        temp_path = "some_temp_path"
        crontab_cmd = "crontab -u {} {}".format("non-root", temp_path)

        with patch.dict(cron.__grains__, {"os_family": "RedHat"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match",
            new=MagicMock(side_effect=[False, True]),
        ), patch(
            "salt.utils.files.fpopen", mock_open()
        ), patch(
            "salt.utils.files.mkstemp", MagicMock(return_value=temp_path)
        ), patch(
            "os.remove", MagicMock()
        ):
            cron._write_cron_lines("non-root", "test 123")
            cron.__salt__["cmd.run_all"].assert_called_with(
                crontab_cmd, python_shell=False
            )

    def test_write_cron_lines_non_root_euid_doesnt_match_user_rh(self):
        """
        Assert that _write_cron_lines() is called with the correct cron command and user
        OS: RedHat. EUID: non-root. EUID doesn't match User.
        Expected to run with runas argument.
        """
        temp_path = "some_temp_path"
        crontab_cmd = "crontab {}".format(temp_path)

        with patch.dict(cron.__grains__, {"os_family": "RedHat"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ), patch(
            "salt.modules.cron._check_instance_uid_match",
            new=MagicMock(return_value=False),
        ), patch(
            "salt.utils.files.fpopen", mock_open()
        ), patch.dict(
            cron.__salt__, {"file.user_to_uid": MagicMock(return_value=1)}
        ), patch(
            "salt.utils.files.mkstemp", MagicMock(return_value=temp_path)
        ), patch(
            "os.remove", MagicMock()
        ):
            cron._write_cron_lines("non-root", "test 123")
            cron.__salt__["cmd.run_all"].assert_called_with(
                crontab_cmd, python_shell=False, runas="non-root"
            )

    def test_write_cron_lines_non_root_aix(self):
        """
        Assert that _write_cron_lines() is called with the correct cron command and user
        OS: AIX. User: non-root.
        Expected to run with runas argument.
        """
        temp_path = "some_temp_path"
        crontab_cmd = "crontab {}".format(temp_path)

        with patch.dict(cron.__grains__, {"os_family": "AIX"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ), patch("salt.utils.files.fpopen", mock_open()), patch.dict(
            cron.__salt__, {"file.user_to_uid": MagicMock(return_value=1)}
        ), patch(
            "salt.utils.files.mkstemp", MagicMock(return_value=temp_path)
        ), patch(
            "os.remove", MagicMock()
        ):
            cron._write_cron_lines("non-root", "test 123")
            cron.__salt__["cmd.run_all"].assert_called_with(
                crontab_cmd, python_shell=False, runas="non-root"
            )

    def test_write_cron_lines_non_root_solaris(self):
        """
        Assert that _write_cron_lines() is called with the correct cron command and user
        OS: Solaris. User: non-root.
        Expected to run with runas argument.
        """
        temp_path = "some_temp_path"
        crontab_cmd = "crontab {}".format(temp_path)

        with patch.dict(cron.__grains__, {"os_family": "Solaris"}), patch.dict(
            cron.__salt__, {"cmd.run_all": MagicMock()}
        ), patch("salt.utils.files.fpopen", mock_open()), patch.dict(
            cron.__salt__, {"file.user_to_uid": MagicMock(return_value=1)}
        ), patch(
            "salt.utils.files.mkstemp", MagicMock(return_value=temp_path)
        ), patch(
            "os.remove", MagicMock()
        ):
            cron._write_cron_lines("non-root", "test 123")
            cron.__salt__["cmd.run_all"].assert_called_with(
                crontab_cmd, python_shell=False, runas="non-root"
            )



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.cassandra_mod
"""

import pytest

import salt.modules.cassandra_mod as cassandra
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {cassandra: {}}


def test_compactionstats():
    """
    Test for Return compactionstats info
    """
    mock = MagicMock(return_value="A")
    with patch.object(cassandra, "_nodetool", mock):
        assert cassandra.compactionstats() == "A"


def test_version():
    """
    Test for Return the cassandra version
    """
    mock = MagicMock(return_value="A")
    with patch.object(cassandra, "_nodetool", mock):
        assert cassandra.version() == "A"


def test_netstats():
    """
    Test for Return netstats info
    """
    mock = MagicMock(return_value="A")
    with patch.object(cassandra, "_nodetool", mock):
        assert cassandra.netstats() == "A"


def test_tpstats():
    """
    Test for Return tpstats info
    """
    mock = MagicMock(return_value="A")
    with patch.object(cassandra, "_nodetool", mock):
        assert cassandra.tpstats() == "A"


def test_info():
    """
    Test for Return cassandra node info
    """
    mock = MagicMock(return_value="A")
    with patch.object(cassandra, "_nodetool", mock):
        assert cassandra.info() == "A"


def test_ring():
    """
    Test for Return ring info
    """
    mock = MagicMock(return_value="A")
    with patch.object(cassandra, "_nodetool", mock):
        assert cassandra.ring() == "A"


def test_keyspaces():
    """
    Test for Return existing keyspaces
    """
    mock_keyspaces = ["A", "B", "C", "D"]

    class MockSystemManager:
        def list_keyspaces(self):
            return mock_keyspaces

    mock_sys_mgr = MagicMock(return_value=MockSystemManager())

    with patch.object(cassandra, "_sys_mgr", mock_sys_mgr):
        assert cassandra.keyspaces() == mock_keyspaces


def test_column_families():
    """
    Test for Return existing column families for all keyspaces
    """
    mock_keyspaces = ["A", "B"]

    class MockSystemManager:
        def list_keyspaces(self):
            return mock_keyspaces

        def get_keyspace_column_families(self, keyspace):
            if keyspace == "A":
                return {"a": "saltines", "b": "biscuits"}
            if keyspace == "B":
                return {"c": "cheese", "d": "crackers"}

    mock_sys_mgr = MagicMock(return_value=MockSystemManager())

    with patch.object(cassandra, "_sys_mgr", mock_sys_mgr):
        assert cassandra.column_families("Z") is None
        assert cassandra.column_families("A") == ["a", "b"]
        assert cassandra.column_families() == {"A": ["a", "b"], "B": ["c", "d"]}


def test_column_family_definition():
    """
    Test for Return a dictionary of column family definitions for the given
    keyspace/column_family
    """

    class MockSystemManager:
        def get_keyspace_column_families(self, keyspace):
            if keyspace == "A":
                return {"a": object, "b": object}
            if keyspace == "B":
                raise Exception

    mock_sys_mgr = MagicMock(return_value=MockSystemManager())

    with patch.object(cassandra, "_sys_mgr", mock_sys_mgr):
        assert cassandra.column_family_definition("A", "a") == vars(object)
        assert cassandra.column_family_definition("B", "a") is None



import pytest

import salt.states.macpackage as macpackage
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {macpackage: {}}


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_pkg():
    """
    Test installing a PKG file
    """
    expected = {
        "changes": {"installed": ["some.other.id"]},
        "comment": "/path/to/file.pkg installed",
        "name": "/path/to/file.pkg",
        "result": True,
    }

    installed_mock = MagicMock(return_value=["com.apple.id"])
    get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
    install_mock = MagicMock(return_value={"retcode": 0})

    with patch.dict(
        macpackage.__salt__,
        {
            "macpackage.installed_pkgs": installed_mock,
            "macpackage.get_pkg_id": get_pkg_id_mock,
            "macpackage.install": install_mock,
        },
    ):
        out = macpackage.installed("/path/to/file.pkg")
        installed_mock.assert_called_once_with()
        get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
        install_mock.assert_called_once_with(
            "/path/to/file.pkg", "LocalSystem", False, False
        )
        assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_pkg_exists():
    """
    Test installing a PKG file where it's already installed
    """
    expected = {
        "changes": {},
        "comment": "",
        "name": "/path/to/file.pkg",
        "result": True,
    }

    installed_mock = MagicMock(return_value=["com.apple.id", "some.other.id"])
    get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
    install_mock = MagicMock(return_value={"retcode": 0})

    with patch.dict(
        macpackage.__salt__,
        {
            "macpackage.installed_pkgs": installed_mock,
            "macpackage.get_pkg_id": get_pkg_id_mock,
            "macpackage.install": install_mock,
        },
    ):
        out = macpackage.installed("/path/to/file.pkg")
        installed_mock.assert_called_once_with()
        get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
        assert not install_mock.called
        assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_pkg_version_succeeds():
    """
    Test installing a PKG file where the version number matches the current installed version
    """
    expected = {
        "changes": {},
        "comment": "Version already matches .*5\\.1\\.[0-9]",
        "name": "/path/to/file.pkg",
        "result": True,
    }

    installed_mock = MagicMock(return_value=["com.apple.id", "some.other.id"])
    get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
    install_mock = MagicMock(return_value={"retcode": 0})
    cmd_mock = MagicMock(return_value="Version of this: 5.1.9")

    with patch.dict(
        macpackage.__salt__,
        {
            "macpackage.installed_pkgs": installed_mock,
            "macpackage.get_pkg_id": get_pkg_id_mock,
            "macpackage.install": install_mock,
            "cmd.run": cmd_mock,
        },
    ):
        out = macpackage.installed(
            "/path/to/file.pkg",
            version_check=r"/usr/bin/runme --version=.*5\.1\.[0-9]",
        )
        cmd_mock.assert_called_once_with(
            "/usr/bin/runme --version", output_loglevel="quiet", ignore_retcode=True
        )
        assert not installed_mock.called
        assert not get_pkg_id_mock.called
        assert not install_mock.called
        assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_pkg_version_fails():
    """
    Test installing a PKG file where the version number if different from the expected one
    """
    expected = {
        "changes": {"installed": ["some.other.id"]},
        "comment": (
            "Version Version of this: 1.8.9 doesn't match .*5\\.1\\.[0-9]."
            " /path/to/file.pkg installed"
        ),
        "name": "/path/to/file.pkg",
        "result": True,
    }

    installed_mock = MagicMock(return_value=["com.apple.id"])
    get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
    install_mock = MagicMock(return_value={"retcode": 0})
    cmd_mock = MagicMock(return_value="Version of this: 1.8.9")

    with patch.dict(
        macpackage.__salt__,
        {
            "macpackage.installed_pkgs": installed_mock,
            "macpackage.get_pkg_id": get_pkg_id_mock,
            "macpackage.install": install_mock,
            "cmd.run": cmd_mock,
        },
    ):
        out = macpackage.installed(
            "/path/to/file.pkg",
            version_check=r"/usr/bin/runme --version=.*5\.1\.[0-9]",
        )
        cmd_mock.assert_called_once_with(
            "/usr/bin/runme --version", output_loglevel="quiet", ignore_retcode=True
        )
        installed_mock.assert_called_once_with()
        get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
        install_mock.assert_called_once_with(
            "/path/to/file.pkg", "LocalSystem", False, False
        )
        assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_dmg():
    """
    Test installing a DMG file
    """
    expected = {
        "changes": {"installed": ["some.other.id"]},
        "comment": "/path/to/file.dmg installed",
        "name": "/path/to/file.dmg",
        "result": True,
    }

    mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
    unmount_mock = MagicMock()
    installed_mock = MagicMock(return_value=["com.apple.id"])
    get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
    install_mock = MagicMock(return_value={"retcode": 0})

    with patch.dict(
        macpackage.__salt__,
        {
            "macpackage.mount": mount_mock,
            "macpackage.unmount": unmount_mock,
            "macpackage.installed_pkgs": installed_mock,
            "macpackage.get_pkg_id": get_pkg_id_mock,
            "macpackage.install": install_mock,
        },
    ):
        out = macpackage.installed("/path/to/file.dmg", dmg=True)
        mount_mock.assert_called_once_with("/path/to/file.dmg")
        unmount_mock.assert_called_once_with("/tmp/dmg-X")
        installed_mock.assert_called_once_with()
        get_pkg_id_mock.assert_called_once_with("/tmp/dmg-X/*.pkg")
        install_mock.assert_called_once_with(
            "/tmp/dmg-X/*.pkg", "LocalSystem", False, False
        )
        assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_dmg_exists():
    """
    Test installing a DMG file when the package already exists
    """
    expected = {
        "changes": {},
        "comment": "",
        "name": "/path/to/file.dmg",
        "result": True,
    }

    mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
    unmount_mock = MagicMock()
    installed_mock = MagicMock(return_value=["com.apple.id", "some.other.id"])
    get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
    install_mock = MagicMock(return_value={"retcode": 0})

    with patch.dict(
        macpackage.__salt__,
        {
            "macpackage.mount": mount_mock,
            "macpackage.unmount": unmount_mock,
            "macpackage.installed_pkgs": installed_mock,
            "macpackage.get_pkg_id": get_pkg_id_mock,
            "macpackage.install": install_mock,
        },
    ):
        out = macpackage.installed("/path/to/file.dmg", dmg=True)
        mount_mock.assert_called_once_with("/path/to/file.dmg")
        unmount_mock.assert_called_once_with("/tmp/dmg-X")
        installed_mock.assert_called_once_with()
        get_pkg_id_mock.assert_called_once_with("/tmp/dmg-X/*.pkg")
        assert not install_mock.called
        assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_app():
    """
    Test installing an APP file
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {"installed": ["file.app"]},
            "comment": "file.app installed",
            "name": "/path/to/file.app",
            "result": True,
        }

        install_mock = MagicMock()
        exists_mock.return_value = False

        with patch.dict(macpackage.__salt__, {"macpackage.install_app": install_mock}):
            out = macpackage.installed("/path/to/file.app", app=True)

            install_mock.assert_called_once_with("/path/to/file.app", "/Applications/")
            assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_app_exists():
    """
    Test installing an APP file that already exists
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {},
            "comment": "",
            "name": "/path/to/file.app",
            "result": True,
        }

        install_mock = MagicMock()
        exists_mock.return_value = True

        with patch.dict(macpackage.__salt__, {"macpackage.install_app": install_mock}):
            out = macpackage.installed("/path/to/file.app", app=True)

            assert not install_mock.called
            assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_app_dmg():
    """
    Test installing an APP file contained in a DMG file
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {"installed": ["file.app"]},
            "comment": "file.app installed",
            "name": "/path/to/file.dmg",
            "result": True,
        }

        install_mock = MagicMock()
        mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
        unmount_mock = MagicMock()
        cmd_mock = MagicMock(return_value="file.app")
        exists_mock.return_value = False

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.install_app": install_mock,
                "macpackage.mount": mount_mock,
                "macpackage.unmount": unmount_mock,
                "cmd.run": cmd_mock,
            },
        ):
            out = macpackage.installed("/path/to/file.dmg", app=True, dmg=True)

            mount_mock.assert_called_once_with("/path/to/file.dmg")
            unmount_mock.assert_called_once_with("/tmp/dmg-X")
            cmd_mock.assert_called_once_with(
                "ls -d *.app", python_shell=True, cwd="/tmp/dmg-X"
            )
            install_mock.assert_called_once_with(
                "/tmp/dmg-X/file.app", "/Applications/"
            )
            assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_app_dmg_exists():
    """
    Test installing an APP file contained in a DMG file where the file exists
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {},
            "comment": "",
            "name": "/path/to/file.dmg",
            "result": True,
        }

        install_mock = MagicMock()
        mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
        unmount_mock = MagicMock()
        cmd_mock = MagicMock(return_value="file.app")
        exists_mock.return_value = True

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.install_app": install_mock,
                "macpackage.mount": mount_mock,
                "macpackage.unmount": unmount_mock,
                "cmd.run": cmd_mock,
            },
        ):
            out = macpackage.installed("/path/to/file.dmg", app=True, dmg=True)

            mount_mock.assert_called_once_with("/path/to/file.dmg")
            unmount_mock.assert_called_once_with("/tmp/dmg-X")
            cmd_mock.assert_called_once_with(
                "ls -d *.app", python_shell=True, cwd="/tmp/dmg-X"
            )
            assert not install_mock.called
            assert out == expected


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_installed_pkg_only_if_pass():
    """
    Test installing a PKG file where the only if call passes
    """
    expected = {
        "changes": {"installed": ["some.other.id"]},
        "comment": "/path/to/file.pkg installed",
        "name": "/path/to/file.pkg",
        "result": True,
    }

    installed_mock = MagicMock(return_value=["com.apple.id"])
    get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
    install_mock = MagicMock(return_value={"retcode": 0})
    cmd_mock = MagicMock(return_value=0)

    with patch.dict(
        macpackage.__salt__,
        {
            "macpackage.installed_pkgs": installed_mock,
            "macpackage.get_pkg_id": get_pkg_id_mock,
            "macpackage.install": install_mock,
            "cmd.retcode": cmd_mock,
        },
    ):
        out = macpackage.installed("/path/to/file.pkg")
        installed_mock.assert_called_once_with()
        get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
        install_mock.assert_called_once_with(
            "/path/to/file.pkg", "LocalSystem", False, False
        )
        assert out == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.pecl as pecl
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pecl: {}}


def test_installed():
    """
    Test to make sure that a pecl extension is installed.
    """
    name = "mongo"
    ver = "1.0.1"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock_lst = MagicMock(return_value={name: "stable"})
    mock_t = MagicMock(return_value=True)
    with patch.dict(pecl.__salt__, {"pecl.list": mock_lst, "pecl.install": mock_t}):
        comt = "Pecl extension {} is already installed.".format(name)
        ret.update({"comment": comt, "result": True})
        assert pecl.installed(name) == ret

        with patch.dict(pecl.__opts__, {"test": True}):
            comt = "Pecl extension mongo-1.0.1 would have been installed"
            ret.update({"comment": comt, "result": None})
            assert pecl.installed(name, version=ver) == ret

        with patch.dict(pecl.__opts__, {"test": False}):
            comt = "Pecl extension mongo-1.0.1 was successfully installed"
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "changes": {"mongo-1.0.1": "Installed"},
                }
            )
            assert pecl.installed(name, version=ver) == ret


def test_removed():
    """
    Test to make sure that a pecl extension is not installed.
    """
    name = "mongo"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock_lst = MagicMock(side_effect=[{}, {name: "stable"}, {name: "stable"}])
    mock_t = MagicMock(return_value=True)
    with patch.dict(pecl.__salt__, {"pecl.list": mock_lst, "pecl.uninstall": mock_t}):
        comt = "Pecl extension {} is not installed.".format(name)
        ret.update({"comment": comt, "result": True})
        assert pecl.removed(name) == ret

        with patch.dict(pecl.__opts__, {"test": True}):
            comt = "Pecl extension mongo would have been removed"
            ret.update({"comment": comt, "result": None})
            assert pecl.removed(name) == ret

        with patch.dict(pecl.__opts__, {"test": False}):
            comt = "Pecl extension mongo was successfully removed."
            ret.update({"comment": comt, "result": True, "changes": {name: "Removed"}})
            assert pecl.removed(name) == ret



"""
Utilities for managing Debian preseed

.. versionadded:: 2015.8.0
"""

import shlex

import salt.utils.files
import salt.utils.stringutils
import salt.utils.yaml


def mksls(src, dst=None):
    """
    Convert a preseed file to an SLS file
    """
    ps_opts = {}
    with salt.utils.files.fopen(src, "r") as fh_:
        for line in fh_:
            line = salt.utils.stringutils.to_unicode(line)
            if line.startswith("#"):
                continue
            if not line.strip():
                continue

            comps = shlex.split(line)
            if comps[0] not in ps_opts.keys():
                ps_opts[comps[0]] = {}
            cmds = comps[1].split("/")

            pointer = ps_opts[comps[0]]
            for cmd in cmds:
                pointer = pointer.setdefault(cmd, {})

            pointer["type"] = comps[2]
            if len(comps) > 3:
                pointer["argument"] = comps[3]

    sls = {}

    # Set language
    # ( This looks like it maps to something else )
    sls[ps_opts["d-i"]["languagechooser"]["language-name-fb"]["argument"]] = {
        "locale": ["system"]
    }

    # Set keyboard
    # ( This looks like it maps to something else )
    sls[ps_opts["d-i"]["kbd-chooser"]["method"]["argument"]] = {"keyboard": ["system"]}

    # Set timezone
    timezone = ps_opts["d-i"]["time"]["zone"]["argument"]
    sls[timezone] = {"timezone": ["system"]}
    if ps_opts["d-i"]["tzconfig"]["gmt"]["argument"] == "true":
        sls[timezone]["timezone"].append("utc")

    # Set network
    if "netcfg" in ps_opts["d-i"].keys():
        iface = ps_opts["d-i"]["netcfg"]["choose_interface"]["argument"]
        sls[iface] = {}
        sls[iface]["enabled"] = True
        if ps_opts["d-i"]["netcfg"]["confirm_static"] == "true":
            sls[iface]["proto"] = "static"
        elif ps_opts["d-i"]["netcfg"]["disable_dhcp"] == "false":
            sls[iface]["proto"] = "dhcp"
        sls[iface]["netmask"] = ps_opts["d-i"]["netcfg"]["get_netmask"]["argument"]
        sls[iface]["domain"] = ps_opts["d-i"]["netcfg"]["get_domain"]["argument"]
        sls[iface]["gateway"] = ps_opts["d-i"]["netcfg"]["get_gateway"]["argument"]
        sls[iface]["hostname"] = ps_opts["d-i"]["netcfg"]["get_hostname"]["argument"]
        sls[iface]["ipaddress"] = ps_opts["d-i"]["netcfg"]["get_ipaddress"]["argument"]
        sls[iface]["nameservers"] = ps_opts["d-i"]["netcfg"]["get_nameservers"][
            "argument"
        ]

    if dst is not None:
        with salt.utils.files.fopen(dst, "w") as fh_:
            salt.utils.yaml.safe_dump(sls, fh_, default_flow_style=False)
    else:
        return salt.utils.yaml.safe_dump(sls, default_flow_style=False)
